#!/usr/bin/env python# -*- coding: utf-8 -*-
r"""
@DATE: 2023-11-30 21:20:13
@File: swanlab\env.py
@IDE: vscode
@Description:
    swanlab全局共用环境变量(运行时环境变量)
    除了utils和error模块，其他模块都可以使用这个模块
"""
import enum
import io
import netrc
import os
import re
import sys
from typing import List
from urllib.parse import urlparse

import swanlab.toolkit as E
from swanlab.toolkit import SwanLabSharedEnv


# ---------------------------------- 环境变量枚举类 ----------------------------------


class SwanLabEnv(enum.Enum):
    """
    swanlab环境变量枚举类，包含swankit的共享环境变量
    """

    SWANLAB_FOLDER = SwanLabSharedEnv.SWANLAB_FOLDER.value
    """
    swanlab全局文件夹保存的路径，默认为用户主目录下的.swanlab文件夹
    """
    SWANLOG_FOLDER = SwanLabSharedEnv.SWANLOG_FOLDER.value
    """
    swanlab解析日志文件保存的路径，默认为当前运行目录的swanlog文件夹
    """
    MODE = SwanLabSharedEnv.SWANLAB_MODE.value
    """
    swanlab的解析模式，涉及操作员注册的回调，目前有四种：local、cloud、disabled、offline，默认为cloud
    大小写敏感
    """
    SWANBOARD_PROT = "SWANLAB_BOARD_PORT"
    """
    cli swanboard 服务端口
    """
    SWANBOARD_HOST = "SWANLAB_BOARD_HOST"
    """
    cli swanboard 服务地址
    """
    WEB_HOST = "SWANLAB_WEB_HOST"
    """
    swanlab云端环境的web地址
    """
    API_HOST = "SWANLAB_API_HOST"
    """
    swanlab云端环境的api地址
    """
    API_KEY = "SWANLAB_API_KEY"
    """
    云端api key，登录时会首先查找此环境变量，如果不存在，判断用户是否已登录，未登录则进入登录流程

    * 如果login接口传入字符串，此环境变量无效，此时相当于绕过 get_key 接口
    * 如果用户已登录，此环境变量的优先级高于本地存储登录信息
    """
    WORKSPACE = "SWANLAB_WORKSPACE"
    """
    swanlab的工作空间，默认为当前登录用户
    """
    PROJ_NAME = "SWANLAB_PROJ_NAME"
    """
    swanlab的项目名称
    """
    EXP_NAME = "SWANLAB_EXP_NAME"
    """
    swanlab的实验名称
    """
    RUN_ID = "SWANLAB_RUN_ID"
    """
    swanlab 实验运行 id，resume 时使用
    """
    RESUME = "SWANLAB_RESUME"

    RUNTIME = "SWANLAB_RUNTIME"
    """
    swanlab的运行时环境，"user" "develop" "test" "test-no-cloud"
    """
    WEBHOOK = "SWANLAB_WEBHOOK"
    """
    webhook地址。swanlab初始化完毕时，如果此环境变量存在，会调用此地址，发送消息。
    """

    @staticmethod
    def is_hostname(value: str) -> bool:
        """
        判断是否为合法的主机名（支持 http/https 协议和端口号）
        :param value: 待判断的字符串
        :return: 是否为合法的主机名
        """
        # 解析 URL 以提取主机部分
        parsed_url = urlparse(value)
        if parsed_url.scheme in ["http", "https"]:
            value = parsed_url.hostname  # 只取域名部分

        # 处理 IP 地址
        if re.fullmatch(r'((25[0-5]|2[0-4][0-9]|1?[0-9]{1,2})(\.|$)){4}', value):
            return True

        # 处理域名（去掉端口）
        if re.fullmatch(r'^(?!-)([a-zA-Z0-9-]{1,63}\.)+[a-zA-Z]{2,63}$', value):
            return True

        return False

    @classmethod
    def set_default(cls):
        """
        设置默认的环境变量值，如果netrc文件存在，使用netrc中第一个machine的信息
        """
        # 从netrc文件中读取host信息
        path = os.path.join(get_save_dir(), ".netrc")
        web_host = "https://swanlab.cn"
        api_host = "https://api.swanlab.cn/api"
        if os.path.exists(path):
            nrc = netrc.netrc(path)
            for host, info in nrc.hosts.items():
                web_host = info[0] if cls.is_hostname(info[0]) else host.rstrip("/api")
                api_host = host.rstrip("/api") + "/api"
                break

        envs = {
            cls.WEB_HOST.value: web_host,
            cls.API_HOST.value: api_host,
            cls.RUNTIME.value: "user",
        }
        for k, v in envs.items():
            os.environ.setdefault(k, v)

    @classmethod
    def check(cls):
        """
        检查环境变量的值是否为预期值中的一个
        :raises ValueError: 如果环境变量的值不在预期值中
        """
        envs = {
            cls.MODE.value: ["local", "cloud", "disabled", "offline"],
            cls.RUNTIME.value: ["user", "develop", "test", "test-no-cloud"],
        }
        for k, vs in envs.items():
            if k in os.environ and os.environ[k] not in vs:
                raise ValueError(f"Unknown value for {k}: {os.environ[k]}")

    @classmethod
    def list(cls) -> List[str]:
        """
        获取所有的枚举值
        :return: 所有的枚举值
        """
        return [item.value for item in cls]


# ---------------------------------- API ----------------------------------

is_windows = E.is_windows

get_mode = E.get_mode

get_swanlog_dir = E.get_swanlog_dir

get_save_dir = E.get_save_dir


def in_jupyter() -> bool:
    """
    用于检测是否在 notebook jupyter 中运行
    :return: bool 是否在 notebook 中运行
    """
    try:
        # notebook 中会有 __IPYTHON__，而正常环境没有定义，所以 try
        # 'type: ignore': 可以让 pylance 忽略对变量定义的检查
        _ = __IPYTHON__  # type: ignore
        return True
    except NameError:
        return False


def is_interactive():
    """
    是否为可交互式环境（输入连接tty设备）
    特殊的环境：jupyter notebook
    """
    try:
        fd = sys.stdin.fileno()
        return os.isatty(fd) or in_jupyter()
    # 当使用capsys、capfd或monkeypatch等fixture来捕获或修改标准输入输出时，会抛出io.UnsupportedOperation
    # 多为测试情况，可交互
    except io.UnsupportedOperation:
        return True



"""
Docs: https://docs.swanlab.cn/guide_cloud/integration/integration-fastai.html
"""

try:
    from fastai.learner import Callback
    from fastcore.basics import store_attr, detuplify, ignore_exceptions
    from fastai.callback.hook import total_params
except ImportError:
    raise RuntimeError(
        "This module requires `fastai` to be installed. " "Please install it with command: \n pip install fastai"
    )

from typing import Optional, Any, Dict, List
import swanlab
from swanlab.log import swanlog as swl


class SwanLabCallback(Callback):
    def __init__(
        self,
        project: Optional[str] = None,
        experiment_name: Optional[str] = None,
        description: Optional[str] = None,
        workspace: Optional[str] = None,
        config: Optional[dict] = None,
        mode: Optional[str] = None,
        logdir: Optional[str] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        store_attr()
        self._experiment = swanlab

        self.project = project
        self.experiment_name = experiment_name
        self.workspace = workspace
        self.config = config
        self.description = description
        self.mode = mode
        self.logdir = logdir
        self.train_suffix = "train"
        self.summary_suffix = "summary"
        
        self.tags = tags or []
        self.tags.append("fastai") if "fastai" not in self.tags else None
    
    def update_config(self, config: Dict[str, Any]):
        swanlab.config.update(config)

    def setup_swanlab(self):
        swanlab.config["FRAMEWORK"] = "fastai"
        if self._experiment.get_run() is None:
            self._experiment.init(
                project=self.project,
                workspace=self.workspace,
                experiment_name=self.experiment_name,
                description=self.description,
                config=self.config,
                mode=self.mode,
                logdir=self.logdir,
                tags=self.tags,
            )

    def before_fit(self):
        # print("=" * 10 + "before_fit" + "=" * 10)
        self.setup_swanlab()
        configs_log = self.gather_args()
        formatted_config = SwanLabCallback.format_config(configs_log)
        self._experiment.config.update(formatted_config)
        self._swanlab_step = 0

    def after_batch(self):
        if self.training:
            self._swanlab_step += 1
            swanlab.log(
                {f"{self.train_suffix}/loss": self.loss.item(), f"{self.train_suffix}/step": self._swanlab_step},
                step=self._swanlab_step,
            )
            for i, h in enumerate(self.opt.hypers):
                for k, v in h.items():
                    swanlab.log({f"{self.train_suffix}/{k}_{i}": v}, self._swanlab_step)

    def before_epoch(self):
        # print("=" * 10 + "before_epoch" + "=" * 10)
        for metric in self.metrics:
            metric.reset()

    def after_epoch(self):
        # print("=" * 10 + "after_epoch" + "=" * 10)
        for name, value in zip(self.recorder.metric_names, self.recorder.log):
            if value is not None:
                swanlab.log({f"{self.summary_suffix}/{name}": value})

    def __del__(self):
        # 如果实验已经结束，且实验状态为0，即RUNNING状态，则关闭实验
        if self._experiment.Run.get_state().value == 0:
            swanlab.finish()

    def gather_args(self):
        "Gather config parameters accessible to the learner"
        cb_args = {f"{cb}": getattr(cb, "__stored_args__", True) for cb in self.cbs if cb != self}
        args = {"Learner": self.learn, **cb_args}
        try:
            n_inp = self.dls.train.n_inp
            args["n_inp"] = n_inp
            xb = self.dls.valid.one_batch()[:n_inp]
            args.update(
                {f"input {n+1} dim {i+1}": d for n in range(n_inp) for i, d in enumerate(list(detuplify(xb[n]).shape))}
            )
        except Exception:
            swl.warning("Failed to gather input dimensions")
        with ignore_exceptions():
            args["batch_size"] = self.dls.bs
            args["batch_per_epoch"] = len(self.dls.train)
            args["model_parameters"] = total_params(self.model)[0]
            args["device"] = self.dls.device.type
            args["frozen"] = bool(self.opt.frozen_idx)
            args["frozen_idx"] = self.opt.frozen_idx
            args["dataset/tfms"] = f"{self.dls.dataset.tfms}"
            args["dls/after_item"] = f"{self.dls.after_item}"
            args["dls/before_batch"] = f"{self.dls.before_batch}"
            args["dls/after_batch"] = f"{self.dls.after_batch}"
        return args

    @classmethod
    def format_config(cls, config):
        "Format config parameters for logging"
        for key, value in config.items():
            if isinstance(value, dict):
                config[key] = SwanLabCallback.format_config(value)
            else:
                config[key] = SwanLabCallback.format_config_value(value)
        return config

    @classmethod
    def format_config_value(cls, value):
        if isinstance(value, list):
            return [SwanLabCallback.format_config_value(item) for item in value]
        elif hasattr(value, "__stored_args__"):
            return {**SwanLabCallback.format_config(value.__stored_args__), "_name": value}
        return value



"""
@author: cunyue
@file: __init__.py
@time: 2025/6/5 14:03
@description: 同步本地数据到云端
"""

import click

from swanlab.core_python import create_client, auth
from swanlab.error import KeyFileError
from swanlab.package import get_key, HostFormatter
from swanlab.sync import sync as sync_logs


@click.command()
@click.argument(
    "path",
    nargs=-1,
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        resolve_path=True,
    ),
    required=True,
)
@click.option(
    "--api-key",
    "-k",
    default=None,
    type=str,
    help="The API key to use for authentication. If not specified, it will use the default API key from the environment."
    "If specified, it will log in using this API key but will not save the key.",
)
@click.option(
    "--host",
    "-h",
    default=None,
    type=str,
    help="The host to sync the logs to. If not specified, it will use the default host.",
)
@click.option(
    "--workspace",
    "-w",
    default=None,
    type=str,
    help="The workspace to sync the logs to. If not specified, it will use the default workspace.",
)
@click.option(
    "--project",
    "-p",
    default=None,
    type=str,
    help="The project to sync the logs to. If not specified, it will use the default project.",
)
def sync(path, api_key, workspace, project, host):
    """
    Synchronize local logs to the cloud.
    """
    # 1. 创建 http 对象
    # 1.1 检查host是否合法，并格式化，注入到环境变量中
    HostFormatter(host)()
    # 1.2 如果输入了 api-key， 使用此 api-key 登录但不保存数据
    try:
        api_key = get_key() if api_key is None else api_key
    except KeyFileError:
        pass
    for p in path:
        # 1.3 登录，创建 http 对象
        log_info = auth.terminal_login(api_key=api_key, save_key=False)
        create_client(log_info)
        # 2. 同步日志
        sync_logs(p, workspace=workspace, project_name=project, login_required=False, raise_error=len(path) == 1)



from pathlib import Path

import pytest

from swanlab.data.modules.object3d import Model3D


class TestModel3D:
    def test_init_validation(self, test_files_dir):
        """测试初始化验证"""
        glb_path = test_files_dir / "sample.glb"

        # 正确的初始化
        model = Model3D(glb_path)
        assert model.glb_path == glb_path

        # 不存在的文件
        with pytest.raises(FileNotFoundError):
            Model3D(Path("not_exist.glb"))

        # 错误的文件类型
        with pytest.raises(ValueError):
            Model3D(test_files_dir / "wrong.txt")

    def test_from_glb_file(self, test_files_dir):
        """测试从GLB文件创建"""
        glb_path = test_files_dir / "sample.glb"
        model = Model3D.from_glb_file(glb_path, caption="Test")
        assert model.step is None
        assert model.caption == "Test"

    def test_parse(self, test_files_dir):
        """测试解析方法"""
        glb_path = test_files_dir / "sample.glb"
        model = Model3D(glb_path)
        filename, buffer = model.parse()
        assert filename.endswith(".glb")
        # 默认step为None，所以就生成了None
        assert "stepNone" in filename
        assert buffer.getvalue()  # 确保buffer不为空



"""
@author: cunyue
@file: network.py
@time: 2024/12/10 14:30
@description: 网络信息采集
"""

import time
from typing import List

import psutil

from .type import HardwareFuncResult, HardwareCollector, HardwareInfo
from .utils import generate_key, random_index, HardwareConfig


def get_network_info() -> HardwareFuncResult:
    """获取网络信息"""
    return None, NetworkCollector()


class NetworkCollector(HardwareCollector):
    """网络信息采集器"""

    def __init__(self):
        super().__init__()
        self.last_net_io = psutil.net_io_counters()
        self.last_time = time.time()
        network_base_config = HardwareConfig(
            y_range=(0, None),
            chart_name="Network Traffic (KB)",
            chart_index=random_index(),
        ).clone()

        # 网络发送速度
        self.sent_key = generate_key("network.sent")
        self.sent_config = network_base_config.clone(metric_name="sent")

        # 网络接收速度
        self.recv_key = generate_key("network.recv")
        self.recv_config = network_base_config.clone(metric_name="received")

        self.unit = 1024

    def collect(self) -> List[HardwareInfo]:
        current_net_io = psutil.net_io_counters()
        current_time = time.time()  # noqa
        time_diff = current_time - self.last_time

        results = [
            self.get_network_sent_speed(current_net_io.bytes_sent, time_diff),
            self.get_network_recv_speed(current_net_io.bytes_recv, time_diff),
        ]

        # 更新上次的值
        self.last_net_io = current_net_io
        self.last_time = current_time
        return results

    def get_network_sent_speed(self, bytes_sent: int, time_diff: float) -> HardwareInfo:
        """
        获取网络发送速度 (KB/s)
        """
        diff = max(0, bytes_sent - self.last_net_io.bytes_sent)
        return {
            "key": self.sent_key,
            "name": "Network Traffic Sent (KB)",
            "value": self.division_guard(diff, time_diff * self.unit),
            "config": self.sent_config,
        }

    def get_network_recv_speed(self, bytes_recv: int, time_diff: float) -> HardwareInfo:
        """
        获取网络接收速度 (KB/s)
        """
        diff = max(0, bytes_recv - self.last_net_io.bytes_recv)
        return {
            "key": self.recv_key,
            "name": "Network Traffic Received (KB)",
            "value": self.division_guard(diff, time_diff * self.unit),
            "config": self.recv_config,
        }



import argparse
import math
import os
from dataclasses import dataclass

import omegaconf
import pytest
import yaml

from swanlab.data.run.config import SwanLabConfig
from swanlab.data.run.config import parse, Line, RuntimeInfo, MutableMapping
from swanlab.data.run.main import SwanLabRun, get_run, swanlog, get_config
from swanlab.env import SwanLabEnv
from tutils.setup import UseMockRunState


@pytest.fixture(scope="function", autouse=True)
def setup_function():
    """
    在当前测试文件下的每个测试函数执行前后执行
    """
    if get_run() is not None:
        get_run().finish()
    swanlog.disable_log()
    yield
    swanlog.enable_log()


def test_parse():
    """
    测试config.parse函数
    """
    # ---------------------------------- omegaConf ----------------------------------
    config_data = {
        "a": 1,
        "b": "mnist",
        "c/d": [1, 2, 3],
        "e/f/h": {"a": 1, "b": {"c": 2}},
    }
    cfg = omegaconf.OmegaConf.create(config_data)
    config = parse(cfg)
    assert yaml.dump(config) == yaml.dump(config_data)

    # ---------------------------------- mmengine ----------------------------------
    from mmengine.config import Config

    config_data = {
        "a": 1,
        "b": "mnist",
        "c/d": [1, 2, 3],
        "e/f/h": {"a": 1, "b": {"c": 2}},
        "g": True,
    }
    cfg = Config(config_data)
    config = parse(cfg)
    assert yaml.dump(config) == yaml.dump(config_data)
    # ---------------------------------- 自定义继承自MutableMapping的类 ----------------------------------

    class Test(MutableMapping):
        def __init__(self, a, b):
            self.data = {"a": a, "b": b}

        def __setitem__(self, __key, __value):
            self.data[__key] = __value

        def __delitem__(self, __key):
            del self.data[__key]

        def __getitem__(self, __key):
            return self.data.get(__key, None)

        def __len__(self):
            return len(self.data)

        def __iter__(self):
            return iter(self.data)

    config_data = {"a": 1, "b": "mnist", "c/d": [1, 2, 3], "e/f/h": {"a": 1, "b": {"c": 2}}, "test": Test(1, 2)}
    config = parse(config_data)
    assert config["test"]["a"] == 1
    assert config["test"]["b"] == 2
    # ---------------------------------- 包含NaN或者INF的dict对象 ----------------------------------
    config_data = {
        "inf": math.inf,
        "nan": math.nan,
    }
    config = parse(config_data)
    assert config["inf"] == Line.inf
    assert config["nan"] == Line.nan

    # ---------------------------------- dataclass support ----------------------------------
    @dataclass
    class MyData:
        a: int
        b: float

    config_data = MyData(10, 20.0)
    config = parse(config_data)
    assert config["a"] == 10
    assert config["b"] == 20.0
    # ---------------------------------- argparse.Namespace ----------------------------------
    config_data = argparse.Namespace(a=1, b="mnist", c=[1, 2, 3], d={"a": 1, "b": {"c": 2}})
    config = parse(config_data)
    assert yaml.dump(config) == yaml.dump(vars(config_data))


def test_parse_base_class():
    """
    继承自基础的类不能绕过parse函数
    """

    class StrChild(str):
        pass

    value = parse(StrChild("abc"))
    assert value == "abc"
    assert yaml.safe_dump({"value": value})

    class IntChild(int):
        pass

    value = parse(IntChild(1))
    assert value == 1
    assert value != "1"
    assert yaml.safe_dump({"value": value})


class TestSwanLabConfigOperation:
    """
    单独测试TestSwanLabRunConfig这个类
    """

    def test_basic_operation_object(self):
        """
        测试类的基本操作，增删改
        """
        config = SwanLabConfig()
        # ---------------------------------- 对象风格设置 ----------------------------------
        config.a = 1
        assert config.a == 1
        config.a = 2
        assert config.a == 2
        with pytest.raises(AttributeError):
            config.__a = 1
        # 不存在的属性
        with pytest.raises(AttributeError):
            config.b  # noqa
        # 删除属性
        del config.a
        with pytest.raises(AttributeError):
            del config.a  # 重复删除报错
        with pytest.raises(AttributeError):
            config.a  # noqa

    def test_basic_operation_dict(self):
        """
        测试字典的基本操作，增删改
        """
        config = SwanLabConfig()
        # ---------------------------------- 字典风格设置 ----------------------------------
        config["a"] = 1
        assert config["a"] == 1
        config["a"] = 2
        assert config["a"] == 2
        # 删除属性
        del config["a"]
        with pytest.raises(KeyError):
            del config["a"]
        with pytest.raises(KeyError):
            config["a"]  # noqa
        # 字典风格可以设置，读取，删除私有属性
        config["__a"] = 1
        assert config["__a"] == 1
        del config["__a"]
        with pytest.raises(KeyError):
            del config["__a"]  # 重复删除失败
        with pytest.raises(KeyError):
            config["__a"]  # noqa
        # int访问，设置
        config[1] = 1  # noqa
        with pytest.raises(TypeError):
            assert config[1] == 1  # noqa

    def test_dict_iter(self):
        """
        测试字典风格的迭代
        """
        config = SwanLabConfig()
        ll = ["a", "b", "c", "d"]
        for i in ll:
            config[i] = i
        assert set(config) == {"a", "b", "c", "d"}
        index = 0
        # 返回顺序相同
        for key in config:
            assert key == ll[index]
            index += 1

    def test_dict_len(self):
        """
        测试字典风格的长度
        """
        config = SwanLabConfig()
        assert len(config) == 0
        config["a"] = 1
        assert len(config) == 1
        config["b"] = 2
        assert len(config) == 2
        del config["a"]
        assert len(config) == 1
        del config["b"]
        assert len(config) == 0

    def test_func_operation(self):
        """
        测试内置函数操作
        """
        config = SwanLabConfig()
        # ---------------------------------- get ----------------------------------
        a = config.get("a")
        assert a is None
        a = config.get("a", 1)
        assert a == 1
        config["a"] = 5
        a = config.get("a")
        assert a == 5
        # ---------------------------------- set ----------------------------------
        config.set("b", 1)
        assert config["b"] == 1
        config.set("__b", 1)
        assert config["__b"] == 1
        # ---------------------------------- pop ----------------------------------
        config["c"] = 9
        c = config.pop("c")
        assert c == 9
        assert config.pop("d") is None
        # ---------------------------------- clean ----------------------------------
        config["e"] = 1
        config["g"] = 0
        config.clean()
        assert len(config) == 0
        with pytest.raises(KeyError):
            config["e"]  # noqa
        # ---------------------------------- update ----------------------------------
        config["x"] = 1
        config["y"] = 2
        config["z"] = {"a": 1, "b": 2}
        config.update({"x": 2, "y": 3, "z": {"a": 2, "b": 3}})
        assert config["x"] == 2
        assert config["y"] == 3
        assert config["z"] == {"a": 2, "b": 3}
        # update自己
        _config = SwanLabConfig()
        _config.update(config)
        assert _config == config
        # update，argparse.Namespace
        _config = SwanLabConfig()
        _config.update(argparse.Namespace(a=1, b=2))
        assert _config["a"] == 1
        assert _config["b"] == 2
        # update, use kwargs
        _config = SwanLabConfig()
        _config.update(a=2, b=1)
        assert _config["a"] == 2
        assert _config["b"] == 1
        # update bool
        _config = SwanLabConfig()
        _config.update({"a": True, "b": False})
        assert _config["a"] is True
        assert _config["b"] is False


def test_on_setter():
    """
    测试on_setter函数，在设置属性时触发
    """
    num = 1

    def on_setter(_: RuntimeInfo):
        nonlocal num
        num += 1

    config = SwanLabConfig(on_setter=on_setter)

    # ---------------------------------- 对象、字典风格 ----------------------------------

    # 设置触发
    config.a = 1
    assert num == 2
    del config.a
    assert num == 3
    config["b"] = 1
    assert num == 4
    del config["b"]
    assert num == 5

    # 读取不触发
    config.x = 1
    assert num == 6
    _ = config.x
    assert num == 6
    config["y"] = 1
    assert num == 7
    _ = config["y"]
    assert num == 7

    # ---------------------------------- api ----------------------------------

    # 设置触发
    config.set("c", 1)
    assert num == 8
    config.pop("c")
    assert num == 9
    config.update({"d": {}})
    assert num == 10
    # 深层设置无法触发
    config.d["e"] = 1
    assert num == 10

    # 读取不触发
    config.get("f", 1)
    config["g"] = 1
    assert num == 11
    config.get("g")
    assert num == 11

    # ---------------------------------- clean以后再设置无法触发 ----------------------------------

    config.clean()
    config.h = 1
    assert num == 11


class TestSwanLabConfigWithRun:
    """
    测试SwanLabConfig与SwanLabRun的交互
    """

    @staticmethod
    def setup_method():
        os.environ[SwanLabEnv.MODE.value] = "disabled"

    def test_use_dict(self):
        """
        正常流程，输入字典
        """
        with UseMockRunState():
            run = SwanLabRun(
                run_config={
                    "a": 1,
                    "b": "mnist",
                    "c/d": [1, 2, 3],
                    "e/f/h": {"a": 1, "b": {"c": 2}},
                    "g": True,
                }
            )
            config = run.config
            _config = get_config()
            assert config.g is _config.g is True
            assert config["a"] == _config["a"] == 1
            assert config["b"] == _config["b"] == "mnist"
            assert config["c/d"] == _config["c/d"] == [1, 2, 3]

    def test_use_omegaconf(self):
        """
        正常流程，输入OmegaConf
        """
        with UseMockRunState():
            run = SwanLabRun(
                run_config=omegaconf.OmegaConf.create(
                    {
                        "a": 1,
                        "b": "mnist",
                        "c/d": [1, 2, 3],
                        "e/f/h": {"a": 1, "b": {"c": 2}},
                    }
                )
            )
            config = run.config
            _config = get_config()
            assert config["a"] == _config["a"] == 1
            assert config["b"] == _config["b"] == "mnist"
            assert config["c/d"] == _config["c/d"] == [1, 2, 3]

    def test_use_argparse(self):
        """
        正常流程，输入argparse.Namespace
        """
        with UseMockRunState():
            run = SwanLabRun(run_config=argparse.Namespace(a=1, b="mnist", c=[1, 2, 3], d={"a": 1, "b": {"c": 2}}))
            config = run.config
            _config = get_config()
            assert config["a"] == _config["a"] == 1
            assert config["b"] == _config["b"] == "mnist"
            assert config["c"] == _config["c"] == [1, 2, 3]

    def test_use_config(self):
        """
        正常流程，输入SwanLabConfig
        """
        with UseMockRunState():
            run = SwanLabRun(
                run_config=SwanLabConfig(
                    {
                        "a": 1,
                        "b": "mnist",
                        "c": [1, 2, 3],
                        "e/f/h": {"a": 1, "b": {"c": 2}},
                    }
                )
            )
            config = run.config
            _config = get_config()
            assert config["a"] == _config["a"] == 1
            assert config["b"] == _config["b"] == "mnist"
            assert config["c"] == _config["c"] == [1, 2, 3]

    def test_after_finish(self):
        """
        测试在finish之后config的变化
        """
        with UseMockRunState():
            run = SwanLabRun(
                run_config={
                    "a": 1,
                    "b": "mnist",
                    "c/d": [1, 2, 3],
                    "e/f/h": {"a": 1, "b": {"c": 2}},
                }
            )
            run.finish()
            config = run.config
            _config = get_config()
            assert len(config) == 4
            assert len(_config) == 0


class TestFmtConfig:
    def test_fmt_config_empty_dict(self):
        config = {}
        SwanLabConfig.fmt_config(config)
        assert config == {}

    def test_fmt_config_single_key_value(self):
        config = {"key": 123}
        SwanLabConfig.fmt_config(config)
        assert config == {"key": {"value": 123, "desc": "", "sort": 0}}

    def test_fmt_config_multiple_keys(self):
        config = {"a": 1, "b": 2}
        SwanLabConfig.fmt_config(config)
        assert config == {"a": {"value": 1, "desc": "", "sort": 0}, "b": {"value": 2, "desc": "", "sort": 1}}

    def test_fmt_config_non_string_keys(self):
        config = {1: "x", 2: "y"}
        SwanLabConfig.fmt_config(config)
        assert config == {1: {"value": "x", "desc": "", "sort": 0}, 2: {"value": "y", "desc": "", "sort": 1}}

    def test_fmt_config_nested_dict_value(self):
        config = {"outer": {"inner": 42}}
        SwanLabConfig.fmt_config(config)
        assert config == {"outer": {"value": {"inner": 42}, "desc": "", "sort": 0}}


def test_revert_config_removes_desc_and_sort_fields():
    config = {
        "key1": {"value": "val1", "desc": "description1", "sort": 2},
        "key2": {"value": "val2", "desc": "description2", "sort": 1},
    }
    expected = {"key2": "val2", "key1": "val1"}
    assert SwanLabConfig.revert_config(config) == expected


def test_revert_config_handles_missing_sort_field():
    config = {
        "key1": {"value": "val1", "desc": "description1"},
        "key2": {"value": "val2", "desc": "description2", "sort": 1},
    }
    expected = {"key1": "val1", "key2": "val2"}
    assert SwanLabConfig.revert_config(config) == expected


def test_revert_config_ignores_non_dict_values():
    config = {
        "key1": {"value": "val1", "desc": "description1", "sort": 2},
        "key2": "invalid_value",
    }
    expected = {"key1": "val1"}
    assert SwanLabConfig.revert_config(config) == expected


def test_revert_config_returns_empty_dict_for_empty_input():
    config = {}
    expected = {}
    assert SwanLabConfig.revert_config(config) == expected


def test_revert_config_preserves_order_based_on_sort_field():
    config = {
        "key1": {"value": "val1", "desc": "description1", "sort": 3},
        "key2": {"value": "val2", "desc": "description2", "sort": 1},
        "key3": {"value": "val3", "desc": "description3", "sort": 2},
    }
    expected = {"key2": "val2", "key3": "val3", "key1": "val1"}
    assert SwanLabConfig.revert_config(config) == expected



"""
@DATE: 2024/5/5 20:22
@File: callback_cloud.py
@IDE: pycharm
@Description:
    云端回调
"""

import shutil
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Optional

from rich.status import Status
from rich.text import Text

from swanlab.core_python import auth
from swanlab.data.callbacker.callback import SwanLabRunCallback
from swanlab.env import in_jupyter
from swanlab.log import swanlog
from swanlab.toolkit import (
    RuntimeInfo,
    MetricInfo,
    ColumnInfo,
)
from . import utils as U
from .. import namer as N
from ..run import get_run
from ..run.config import SwanLabConfig
from ..run.metadata.hardware import is_system_key
from ..store import get_run_store, RemoteMetric
from ...core_python import *
from ...log.type import LogData


class CloudPyCallback(SwanLabRunCallback):
    login_info: Optional[auth.LoginInfo] = None

    def __init__(self):
        super().__init__()
        self.executor = ThreadPoolExecutor(max_workers=1)

    def __str__(self):
        return "SwanLabCloudPyCallback"

    @staticmethod
    def _create_client():
        try:
            http = get_client()
        except ValueError:
            swanlog.debug("Login info is None, get login info.")
            login_info = CloudPyCallback.login_info
            if login_info is not None:
                # 如果有登录信息，则使用该信息创建客户端
                http = create_client(login_info)
                CloudPyCallback.login_info = None
            else:
                # 如果没有登录信息，则需要用户登录
                # 但是不保存登录信息到本地
                http = create_client(auth.create_login_info(save=False))
        return http

    @staticmethod
    def _converter_summarise_metric():
        pass

    def on_init(self, *args, **kwargs):
        http = self._create_client()
        # 检测是否有最新的版本
        U.check_latest_version()
        run_store = get_run_store()
        with Status("Creating experiment...", spinner="dots"):
            http.mount_project(run_store.project, run_store.workspace, run_store.visibility)
            exp_count = http.history_exp_count
            run_store.run_name = N.generate_name(exp_count) if run_store.run_name is None else run_store.run_name
            run_store.description = "" if run_store.description is None else run_store.description
            run_store.run_colors = N.generate_colors(exp_count)
            run_store.tags = [] if run_store.tags is None else run_store.tags
            try:
                new = http.mount_exp(
                    exp_name=run_store.run_name,
                    colors=run_store.run_colors,
                    description=run_store.description,
                    tags=run_store.tags,
                    cuid=run_store.run_id,
                    must_exist=run_store.resume == 'must',
                )
                run_store.new = new
            except RuntimeError:
                raise RuntimeError(
                    "When resume=must, the experiment must exist in project {}. Please check your parameters.".format(
                        http.proj.name
                    )
                )
            run_store.run_id = http.exp_id
            # 如果不是新实验，需要获取最新的实验配置和指标数据进行本地同步
            if not new:
                # 1. 解析 config，保存到 run_store.config
                config = http.exp.config
                run_store.config = SwanLabConfig.revert_config(config)
                # 2. 获取最新的指标数据，解析为 run_store.metrics
                # 指标总结数据，{log:[{key: '', step: int}] or None, media: [{key: str, step: int}, ...] or None, scalar: [{key: str, step: int}, ...] or None}
                summaries, _ = http.get(
                    "/house/metrics/summaries/{}/{}".format(
                        http.exp.root_proj_cuid or http.proj.cuid, http.exp.root_exp_cuid or http.exp_id
                    ),
                    {"all": True},
                )
                # 设置历史指标条数
                run_store.log_epoch = summaries.get("log", [{}])[0].get("step", 0)
                # 列响应
                columns_resp, _ = http.get("/experiment/{}/column".format(http.exp_id), {"all": True})
                # 列信息, [{error: dict or None, key:str, type:str, class:str}]
                columns = columns_resp.get("list", [])
                # key -> (column_type, column_class, error, latest step)
                metrics: RemoteMetric = {}
                for column in columns:
                    # 从列信息中获取指标信息
                    key = column["key"]
                    column_type = column["type"]
                    column_class = column["class"]
                    error = column.get("error", None)
                    if column_class == "SYSTEM" and not is_system_key(key):
                        # 只记录 sdk 生成的系统指标
                        continue
                    # 从总结数据中获取最新的 step
                    # 这里需要同时查找 media 和 scalar
                    latest_step = None
                    for scalar_summary in summaries.get("scalar") or []:
                        if scalar_summary["key"] == key:
                            latest_step = scalar_summary["step"]
                            break
                    if latest_step is None:
                        for media_summary in summaries.get("media") or []:
                            if media_summary["key"] == key:
                                latest_step = media_summary["step"]
                                break
                    metrics[key] = (column_type, column_class, error, latest_step)
                run_store.metrics = metrics

    def _terminal_handler(self, log_data: LogData):
        self.porter.trace_log(log_data)

    def on_run(self, *args, **kwargs):
        self.porter.open_for_trace(sync=False)
        # 注册终端代理和系统回调
        self._start_terminal_proxy()
        self._register_sys_callback()
        # 打印实验开始信息，在 cloud 模式下如果没有开启 backup 的话不打印“数据保存在 xxx”的信息
        U.print_train_begin(run_dir=self.run_store.run_dir)
        http = get_client()
        swanlog.info("👋 Hi ", Text(http.username, "bold default"), ",welcome to swanlab!", sep="")
        swanlog.info("Syncing run", Text(self.run_store.run_name, "yellow"), "to the cloud")
        experiment_url = U.print_cloud_web()
        # 在Jupyter Notebook环境下，显示按钮
        if in_jupyter():
            U.show_button_html(experiment_url)

    def on_runtime_info_update(self, r: RuntimeInfo, *args, **kwargs):
        self.porter.trace_runtime_info(r)

    def on_column_create(self, column_info: ColumnInfo, *args, **kwargs):
        self.porter.trace_column(column_info)

    def on_metric_create(self, metric_info: MetricInfo, *args, **kwargs):
        # 有错误就不上传
        if metric_info.error:
            return
        self.porter.trace_metric(metric_info)

    def on_stop(self, error: str = None, *args, **kwargs):
        success = get_run().success
        http = get_client()
        if http.pending:
            swanlog.warning("This run was destroyed but it is pending!")
        # 打印信息
        U.print_cloud_web()
        error_epoch = swanlog.epoch + 1
        self._unregister_sys_callback()
        self.porter.close_trace(success, error=error, epoch=error_epoch)
        # 更新实验状态，在此之后实验会话关闭
        http.update_state(success)
        reset_client()
        if not self.user_settings.backup:
            shutil.rmtree(self.run_store.run_dir, ignore_errors=True)



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Tuple

from swanlab.toolkit import DataSuite as D, MediaBuffer, MediaType


@dataclass()
class Model3D(MediaType):
    """3D model data representation class that handles GLB format files.

    This class provides functionality to handle 3D model data in GLB format for visualization.
    It supports loading GLB files and manages metadata like caption.

    Attributes:
        glb_path: Path to the GLB file
        caption: Optional description text

    Examples:
        >>> # Create from GLB file
        >>> model = Model3D.from_glb_file("model.glb")
        >>>
        >>> # Create with metadata
        >>> model = Model3D.from_glb_file(
        ...     "model.glb",
        ...     caption="My 3D Model"
        ... )
        >>>
        >>> # Use with SwanLab
        >>> import swanlab
        >>> swanlab.log({"model": model})
    """

    glb_path: Path
    step: Optional[int] = None
    caption: Optional[str] = None

    def __post_init__(self):
        """Validate input data after initialization.

        Checks:
            1. File exists
            2. Path points to a file
            3. File has .glb extension

        Raises:
            FileNotFoundError: If GLB file does not exist
            ValueError: If path is not a file or not a GLB file
        """
        if not self.glb_path.exists():
            raise FileNotFoundError(f"GLB file not found: {self.glb_path}")

        if not self.glb_path.is_file():
            raise ValueError(f"Path is not a file: {self.glb_path}")

        if self.glb_path.suffix.lower() != '.glb':
            raise ValueError(f"File must be a GLB file: {self.glb_path}")

    @classmethod
    def from_glb_file(cls, path: Path, *, caption: Optional[str] = None, **kwargs) -> "Model3D":
        """Create Model3D instance from a GLB file.

        This is the main factory method to create Model3D objects. It handles file validation
        and metadata management.

        Args:
            path: Path to the .glb file
            caption: Optional description text for the model
            **kwargs: Additional keyword arguments passed to constructor

        Returns:
            Model3D: A new Model3D instance

        Raises:
            FileNotFoundError: If GLB file does not exist
            ValueError: If file is not a GLB file

        Examples:
            >>> # Basic usage
            >>> model = Model3D.from_glb_file("model.glb")
            >>>
            >>> # With metadata
            >>> model = Model3D.from_glb_file(
            ...     "model.glb",
            ...     caption="My 3D Model"
            ... )
        """
        return cls(path, caption=caption, **kwargs)

    # ---------------------------------- override ----------------------------------

    def parse(self) -> Tuple[str, MediaBuffer]:
        """Convert model to buffer for transmission.

        This method reads the GLB file and prepares it for transmission. It:
        1. Reads the file content into a buffer
        2. Generates a unique hash from the content
        3. Creates a unique filename using the original name, step, and hash

        Returns:
            Tuple[str, MediaBuffer]: (filename, file_content)
            - filename: Generated unique filename
            - file_content: File content in MediaBuffer

        Example:
            >>> model = Model3D.from_glb_file("chair.glb")
            >>> filename, content = model.parse()
            >>> print(filename)
            'chair-step1-a1b2c3d4e5f6g7h8.glb'
        """
        with open(self.glb_path, 'rb') as f:
            buffer = MediaBuffer()
            buffer.write(f.read())

        buffer.seek(0)
        hash_name = D.get_hash_by_bytes(buffer.getvalue())[:16]
        save_name = f"{self.glb_path.stem}-step{self.step}-{hash_name}.glb"
        buffer.seek(0)

        return save_name, buffer

    def get_chart(self) -> MediaType.Chart:
        """Get chart type for visualization.

        Returns:
            MediaType.Chart.OBJECT3D: Indicates this is a 3D object
        """
        return MediaType.Chart.OBJECT3D

    def get_section(self) -> str:
        """Get section name for organization.

        Returns:
            str: Section name "Model3d"
        """
        return "Model3d"

    def get_more(self) -> Optional[Dict[str, str]]:
        """Get additional metadata.

        Returns:
            Optional[Dict[str, str]]: Dictionary containing caption if set
        """
        return {"caption": self.caption} if self.caption else None



"""
@author: cunyue
@file: __init__.py.py
@time: 2025/6/22 19:46
@description: 数据搬运工，在 callbacker 中被实例化，支持：
1. 数据上报（可选）：将数据上报到 SwanLab 服务器
2. 数据备份（必须）：将数据备份到本地文件系统
3. 读取本地备份并上报：读取本地备份文件并上报到 SwanLab 服务器

其中数据上报可选择使用 go 后端或者 python 线程
SwanLab 不生产数据，我们只是 AI 训练的搬运工，祝愿所有训练者跑出预期的 loss :)
"""

import os
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Literal, List, Union, Tuple

import wrapt

from swanlab.core_python import get_client
from swanlab.core_python.uploader.thread import ThreadPool, UploadType
from swanlab.data.store import RunStore, get_run_store, reset_run_store
from swanlab.log.type import LogData
from swanlab.proto.v0 import Log, Header, Project, Experiment, Column, Metric, BaseModel, Runtime, Footer, Media, Scalar
from swanlab.toolkit import MetricInfo, ColumnInfo, RuntimeInfo, create_time
from .datastore import DataStore

__all__ = ['DataPorter']


def traced():
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        if instance is not None and getattr(instance, "_mode") != 1:
            raise RuntimeError("ProtoTransfer has not been started for tracing. Call oopen_for_trace() first.")
        return wrapped(*args, **kwargs)

    return wrapper


def synced():
    """
    标记需要同步上传的函数
    """

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        if instance is not None and getattr(instance, "_mode") != 2:
            raise RuntimeError("ProtoTransfer has not been started for sync upload. Call open_for_sync() first.")
        return wrapped(*args, **kwargs)

    return wrapper


def backup():
    """
    标记需要备份的函数，并将函数的返回结果存储到备份文件中
    """

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        result: Union[List[BaseModel], BaseModel] = wrapped(*args, **kwargs)
        if result is not None:
            try:
                f = getattr(instance, "_f")
                if isinstance(result, list):
                    [f.write(item.to_record()) for item in result]
                else:
                    f.write(result.to_record())
            except ValueError:
                # 写入备份文件失败，可能是因为没有开启备份模式或者备份文件未打开
                # TODO: 记录本地日志
                pass
        return result

    return wrapper


def async_io():
    """
    标记需要在其他线程处理的函数
    """

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        executor: ThreadPoolExecutor = getattr(instance, "_executor", None)
        # 如果没有设置 executor，则直接执行函数
        if executor is None:
            return wrapped(*args, **kwargs)
        # 与 https://github.com/SwanHubX/SwanLab/issues/889 相同的问题
        # 在回调中线程池已经关闭，我们需要在主线程中执行
        if executor._shutdown:
            return wrapped(*args, **kwargs)
        # 否则将函数提交到线程池中执行
        executor.submit(wrapped, *args, **kwargs)

    return wrapper


class DataPorter:
    _instance: Optional['DataPorter'] = None
    _run_store: Optional[RunStore] = None

    def __init__(self):
        # 上传线程池，可以选择不开启
        self._pool: Optional[ThreadPool] = None
        # 数据存储句柄
        self._f = DataStore()
        # 当前模式
        # 0: 不开启任何模式
        # 1: 实验日志跟踪模式
        # 2: 同步上传模式，用于本地日志上报
        self._mode: Literal[0, 1, 2] = 0
        # 工作线程池，开启后一些方法会运行在子线程中
        self._executor: Optional[ThreadPoolExecutor] = None
        self._closed = False

        # ---------------------------------- 同步时用到的参数 ----------------------------------
        self._header: Optional[Header] = None
        self._project: Optional[Project] = None
        self._experiment: Optional[Experiment] = None
        self._logs: List[Log] = []
        self._runtime: Runtime = Runtime(
            conda_filename=None,
            requirements_filename=None,
            metadata_filename=None,
            config_filename=None,
        )
        self._columns: List[Column] = []
        self._scalars: List[Scalar] = []
        self._medias: List[Media] = []
        self._footer: Optional[Footer] = None

    def __new__(cls):
        """
        开启单例模式，确保同时只有一个 ProtoTransfer 实例存在，即代表一个实验会话
        创建会话之前，client 必须存在
        """
        if cls._instance is not None:
            raise RuntimeError("DataPorter instance already exists, cannot create a new one.")
        run_store = get_run_store()
        cls._run_store = run_store
        cls._instance = super(DataPorter, cls).__new__(cls)
        return cls._instance

    def _set_mode(self, mode: Literal[0, 1, 2]):
        """
        设置当前模式，只允许从0设置为其他模式
        """
        assert mode in [0, 1, 2], "Mode must be one of [0, 1, 2]."
        assert self._mode == 0, "DataPorter is already in use, cannot change mode."
        assert self._closed is False, "DataPorter has already rested, cannot change mode."
        self._mode = mode

    def open_for_trace(self, sync: bool = False, backend: Literal['go', 'python', 'none'] = 'python'):
        """
        开启日志传输器以实验日志跟踪，创建数据存储句柄
        此函数在同一实例中只能调用一次，后续调用会抛出异常
        :param sync: 是否同步上传数据，默认为 False，表示异步上传
        :param backend: 上传后端类型，默认为 'python'，可选 'go' 或 'none', none 表示不开启上传
        :raises RuntimeError: 如果已经开启了跟踪模式或者别的模式，则抛出异常
        """
        assert self._run_store.media_dir, "Media directory must be set before creating ProtoTransfer instance"
        assert self._run_store.file_dir, "File directory must be set before creating ProtoTransfer instance"
        assert self._run_store.backup_file, "Backup file must be set before creating ProtoTransfer instance"

        assert self._mode == 0, "DataPorter is already in use, cannot open for trace again."
        assert self._closed is False, "DataPorter has already rested, cannot open for trace again."

        self._f.open_for_write(self._run_store.backup_file)
        if backend == 'python':
            # 检查是否已经创建 client
            try:
                get_client()
            except ValueError:
                raise ValueError("Client not initialized when creating ProtoTransfer instance")
            self._pool = ThreadPool()
        elif backend == 'go':
            raise NotImplementedError("swanlab-core is not ready yet.")
        elif backend == 'none':
            pass
        else:
            raise ValueError(f"Unsupported backend: {backend}")

        if not sync:
            self._executor = ThreadPoolExecutor(max_workers=1)

        # 写入备份文件头
        self._f.write(
            Header.model_validate(
                {
                    "create_time": create_time(),
                    "backup_type": "DEFAULT",
                }
            ).to_record()
        )
        project_name = self._run_store.project
        run_name = self._run_store.run_name
        workspace = self._run_store.workspace
        visibility = self._run_store.visibility
        description = self._run_store.description
        tags = self._run_store.tags
        self._f.write(
            Project.model_validate(
                {
                    "name": project_name,
                    "workspace": workspace,
                    "public": visibility,
                }
            ).to_record()
        )
        self._f.write(
            Experiment.model_validate(
                {
                    "name": run_name,
                    "description": description,
                    "tags": tags,
                }
            ).to_record()
        )
        self._set_mode(1)

    def _publish(self, *args, **kwargs):
        """
        发布数据到上传线程池，如果未开启线程池，则不执行任何操作
        """
        if self._pool is not None:
            self._pool.queue.put(*args, **kwargs)

    @async_io()
    @backup()
    @traced()
    def trace_column(self, data: ColumnInfo) -> BaseModel:
        """
        追踪列数据
        """
        column = Column.from_column_info(data)
        self._publish((UploadType.COLUMN, [column.to_column_model()]))
        return column

    @async_io()
    @backup()
    @traced()
    def trace_metric(self, data: MetricInfo) -> BaseModel:
        """
        追踪指标数据
        """
        assert data.error is None, "MetricInfo must not have error, if it has error, do not upload it."
        if data.column_info.chart_type == data.column_info.chart_type.LINE:
            # 标量
            scalar = Metric.from_metric_info(data)
            self._publish((UploadType.SCALAR_METRIC, [scalar.to_scalar_model()]))
            return scalar
        else:
            # 媒体
            media = Metric.from_metric_info(data)
            # 写入媒体信息
            if data.metric_buffers is not None:
                for i, r in enumerate(data.metric_buffers):
                    if r is None:
                        continue
                    # 组合路径
                    path = os.path.join(data.swanlab_media_dir, data.column_info.kid)
                    os.makedirs(path, exist_ok=True)
                    # 写入数据
                    with open(os.path.join(path, data.metric["data"][i]), "wb") as f:
                        f.write(r.getvalue())
            self._publish((UploadType.MEDIA_METRIC, [media.to_media_model(data.swanlab_media_dir)]))
            return media

    @async_io()
    @backup()
    @traced()
    def trace_runtime_info(self, data: RuntimeInfo) -> BaseModel:
        """
        追踪运行时信息
        """
        runtime = Runtime.from_runtime_info(data)
        # 写入运行时信息到数据存储
        file_dir = self._run_store.file_dir
        if data.requirements is not None:
            data.requirements.write(file_dir)
        if data.metadata is not None:
            data.metadata.write(file_dir)
        if data.config is not None:
            data.config.write(file_dir)
        if data.conda is not None:
            data.conda.write(file_dir)
        self._publish((UploadType.FILE, [runtime.to_file_model(file_dir)]))
        return runtime

    @async_io()
    @backup()
    @traced()
    def trace_log(self, data: LogData) -> List[BaseModel]:
        """
        追踪日志数据，日志数据比较特殊，一次可能好几行
        """
        logs = Log.from_log_data(data)
        self._publish((UploadType.LOG, [log.to_log_model() for log in logs]))
        return logs

    @traced()
    def close_trace(self, success: bool, error: str = None, epoch: int = None):
        """
        停止日志跟踪，清理相关资源
        """
        assert self._closed is False, "DataPorter has already rested, cannot close trace again."
        assert self._mode == 1, "DataPorter is not in trace mode, cannot close trace."

        # 上传错误日志
        if error is not None:
            log = Log.model_validate({"level": "ERROR", "message": error, "create_time": create_time(), "epoch": epoch})
            self._publish((UploadType.LOG, [log.to_log_model()]))
            # 备份日志
            self._f.write(log.to_record())
        # 停止worker
        if self._executor is not None:
            self._executor.shutdown(wait=True)
        # 停止上传线程池
        if self._pool is not None:
            self._pool.finish()
        # 写入结束标志
        footer = Footer.model_validate({"create_time": create_time(), "success": success})
        self._f.write(footer.to_record())
        self._f.ensure_flushed()
        self._f.close()
        self._closed = True
        DataPorter._reset()

    @classmethod
    def _reset(cls):
        """重置单例实例，允许创建新的DataPorter"""
        cls._instance = None
        cls._run_store = None

    def open_for_sync(self, run_dir: str, backend: Literal['python', 'go'] = 'python') -> "DataPorter":
        """
        开启同步模式，此函数应该与 __enter__() 一起使用
        """
        self._run_store.run_dir = run_dir
        assert self._run_store.media_dir, "Media directory must be set before creating ProtoTransfer instance"
        assert self._run_store.file_dir, "File directory must be set before creating ProtoTransfer instance"
        assert self._run_store.backup_file, "Backup file must be set before creating ProtoTransfer instance"
        backup_file = self._run_store.backup_file
        assert os.path.isfile(backup_file), f"Backup file {backup_file} does not exist."
        self._f.open_for_scan(backup_file)
        if backend == 'python':
            self._pool = ThreadPool()
        elif backend == 'go':
            raise NotImplementedError("swanlab-core is not ready yet.")
        else:
            raise ValueError(f"Unsupported backend for sync: {backend}")
        self._set_mode(2)
        return self

    @synced()
    def __enter__(self):
        """
        开启本地数据备份和上传线程池
        使用方式为：
        with DataPorter().open_for_sync(backup_file) as porter:
            porter.parse()
            porter.synchronize()
        """
        assert self._mode == 2, "DataPorter is already in use, cannot open for sync."
        assert self._closed is False, "DataPorter has already rested, cannot open for sync."
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        reset_run_store()
        self._closed = True
        DataPorter._reset()
        # 出现错误则抛出异常
        if exc_type is not None:
            raise exc_val

    @synced()
    def parse(self) -> Tuple[Project, Experiment]:
        """
        解析备份文件中的记录，必须在 open_for_sync() 后调用
        """
        for record in self._f:
            if record is None:
                continue
            self._parse_record(record)
        # 检查是否所有必要的记录都已解析
        assert self._header is not None, "Header not parsed"
        # 检查备份文件
        assert (
            self._header.backup_type == "DEFAULT"
        ), f"Backup type mismatch: {self._header.backup_type}, please update your swanlab package."
        assert self._project is not None, "Project not parsed"
        assert self._experiment is not None, "Experiment not parsed"
        return self._project, self._experiment

    def close(self):
        """
        关闭实例，此函数用于没有开启 trace 和 sync 时关闭实例
        """
        assert self._closed is False, "DataPorter has already rested, cannot close."
        assert self._mode == 0, "DataPorter is in use, cannot simply close."
        DataPorter._reset()

    @synced()
    def synchronize(self) -> bool:
        """
        同步上传数据到 SwanLab 服务器，必须在 open_for_sync() 后调用
        返回最终的实验结果，true 代表实验状态为 success 否则为 false
        """
        assert self._mode == 2, "Must synchronize in sync mode (mode=2)."
        assert self._closed is False, "DataPorter has already rested, cannot synchronize."
        # 同步上传数据到 SwanLab 服务器
        self._publish((UploadType.FILE, [self._runtime.to_file_model(file_dir=self._run_store.file_dir)]))
        self._publish((UploadType.COLUMN, [column.to_column_model() for column in self._columns]))
        self._publish((UploadType.SCALAR_METRIC, [scalar.to_scalar_model() for scalar in self._scalars]))
        self._publish(
            (UploadType.MEDIA_METRIC, [media.to_media_model(self._run_store.media_dir) for media in self._medias])
        )
        self._publish((UploadType.LOG, [log.to_log_model() for log in self._logs]))
        self._pool.finish()
        return self._footer.success if self._footer else False

    def _parse_record(self, data: str):
        assert self._mode == 2, "Must parse records in sync mode (mode=2)."
        assert self._closed is False, "DataPorter has already rested, cannot parse."
        record = BaseModel.from_record(data)
        if isinstance(record, Header):
            assert self._header is None, "Header already parsed"
            self._header = record
            return
        if isinstance(record, Project):
            assert self._project is None, "Project already parsed"
            self._project = record
            return
        if isinstance(record, Experiment):
            assert self._experiment is None, "Experiment already parsed"
            self._experiment = record
            return
        if isinstance(record, Log):
            self._logs.append(record)
            return
        if isinstance(record, Runtime):
            if record.conda_filename is not None:
                self._runtime.conda_filename = record.conda_filename
            if record.requirements_filename is not None:
                self._runtime.requirements_filename = record.requirements_filename
            if record.metadata_filename is not None:
                self._runtime.metadata_filename = record.metadata_filename
            if record.config_filename is not None:
                self._runtime.config_filename = record.config_filename
            return
        if isinstance(record, Column):
            self._columns.append(record)
            return
        if isinstance(record, Scalar):
            self._scalars.append(record)
            return
        if isinstance(record, Media):
            self._medias.append(record)
            return
        if isinstance(record, Footer):
            assert self._footer is None, "Footer already parsed"
            self._footer = record
            return
        raise ValueError(f"Unknown record type: {type(record)}")



"""
@author: cunyue
@file: test_conda.py
@time: 2025/3/2 13:08
@description: 测试conda
"""

import subprocess

import pytest
import yaml

from swanlab.data.run.metadata.conda import get_conda

has_conda = subprocess.run(["conda", "--version"], shell=True, capture_output=True).returncode == 0


@pytest.mark.skipif(not has_conda, reason="conda is not installed")
def test_conda():
    conda_info = get_conda()
    assert isinstance(conda_info, str)
    # 可被yaml解析
    load_conda = yaml.safe_load(conda_info)
    assert load_conda is not None
    assert isinstance(load_conda, dict)


@pytest.mark.skipif(has_conda, reason="conda is installed")
def test_conda_not_installed():
    assert get_conda() is None



"""
@author: cunyue
@file: test_utils.py
@time: 2024/12/9 20:55
@description: 测试硬件信息采集工具
"""

from swanlab.data.run.metadata.hardware.utils import random_index, CpuBaseCollector, generate_key


def test_random_index():
    s = random_index()
    assert len(s) == 8
    assert s.isalnum()
    s = random_index(10)
    assert len(s) == 10
    assert s.isalnum()


def test_generate_key():
    s = generate_key("test")
    assert s == "__swanlab__.test"


def test_cpu_usage():
    c = CpuBaseCollector()
    usage = c.get_cpu_usage()
    assert usage is not None
    assert 0 <= usage["value"] <= 100
    assert usage["key"].endswith("cpu.pct")
    assert usage["name"] == "CPU Utilization (%)"
    assert usage["config"].y_range == (0, 100)
    assert usage["config"].chart_name == "CPU Utilization (%)"
    assert usage["config"].chart_index is None


def test_per_cpu_usage():
    c = CpuBaseCollector()
    usage = c.get_per_cpu_usage()
    assert usage is not None
    for idx, u in enumerate(usage):
        assert 0 <= u["value"] <= 100
        assert u["key"].endswith(f"cpu.{idx}.pct")
        assert u["name"] == f"CPU {idx} Utilization (%)"
        assert u["config"].y_range == (0, 100)
        assert u["config"].chart_name == f"CPU Utilization (per core) (%)"
        # 每个核心的index应该相同,因为必须要放在同一个图表中
        assert u["config"].metric_name == f"CPU {idx}"



from swanlab.proto.common.v1 import common_pb2 as _common_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CollectorUploadRequest(_message.Message):
    __slots__ = ("data",)
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _common_pb2.KeyValueList
    def __init__(self, data: _Optional[_Union[_common_pb2.KeyValueList, _Mapping]] = ...) -> None: ...

class CollectorUploadResponse(_message.Message):
    __slots__ = ("success", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...



import json

import numpy as np
import pytest


@pytest.fixture
def xyz_points():
    """生成示例XYZ点云数据"""
    return np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ],
        dtype=np.float32,
    )


@pytest.fixture
def xyzc_points():
    """生成示例XYZC点云数据"""
    points = np.zeros((4, 4), dtype=np.float32)
    points[:, :3] = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ]
    )
    points[:, 3] = [0, 1, 2, 3]  # 类别标签
    return points


@pytest.fixture
def xyzrgb_points():
    """生成示例XYZRGB点云数据"""
    points = np.zeros((4, 6), dtype=np.float32)
    points[:, :3] = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ]
    )
    points[:, 3:] = np.array(
        [
            [255, 0, 0],
            [0, 255, 0],
            [0, 0, 255],
            [255, 255, 255],
        ]
    )
    return points


@pytest.fixture(scope="session")
def test_files_dir(tmp_path_factory):
    """创建会话级别的测试文件"""
    test_dir = tmp_path_factory.mktemp("test_files")

    # 创建示例GLB文件
    glb_file = test_dir / "sample.glb"
    glb_file.write_bytes(b'dummy glb content')

    # 创建错误类型文件
    txt_file = test_dir / "wrong.txt"
    txt_file.write_text('dummy text content')

    # 创建示例点云文件
    points_data = [
        [0, 0, 0, 255, 0, 0],  # 红色点
        [1, 0, 0, 0, 255, 0],  # 绿色点
        [0, 1, 0, 0, 0, 255],  # 蓝色点
        [0, 0, 1, 255, 255, 255],  # 白色点
    ]
    data = {"points": points_data}
    pts_file = test_dir / "points.swanlab.pts.json"
    pts_file.write_text(json.dumps(data))

    return test_dir



"""
@author: cunyue
@file: ascend.py
@time: 2024/11/26 12:32
@description: 华为昇腾NPU信息采集
"""

import math
import os
import platform
import subprocess
from typing import Optional

from ..type import HardwareCollector as H
from ..type import HardwareConfig, HardwareFuncResult, HardwareInfo, HardwareInfoList
from ..utils import generate_key, random_index


def get_ascend_npu_info() -> HardwareFuncResult:
    """
    获取华为昇腾NPU信息，包括驱动版本、设备信息等
    目前的信息统计粒度只到NPU ID级别，没有到Chip ID级别
    """
    # ascend芯片只支持Linux系统
    if platform.system() != "Linux":
        return None, None
    # /dev目录下没有davinci*设备文件，跳过
    # 其实理论上davinci后接数字，代表此设备id，但是官方文档也没明确写，以防万一还是不这么干了
    if not list(filter(lambda x: x.startswith("davinci"), os.listdir("/dev"))):
        return None, None
    info = {"driver": None, "npu": None, "cann": None}
    collector = None
    try:
        # 获取NPU驱动版本
        info["driver"] = get_version()
        # 获取CANN版本
        info["cann"] = get_cann_version()

        # 获取所有NPU设备ID
        npu_map = map_npu()
        hbm_value: int = 0
        for npu_id in npu_map:
            for chip_id in npu_map[npu_id]:
                chip_info = npu_map[npu_id][chip_id]
                usage = get_chip_usage(npu_id, chip_id)
                if info["npu"] is None:
                    info["npu"] = {}
                if npu_id not in info["npu"]:
                    info["npu"][npu_id] = {}
                info["npu"][npu_id][chip_id] = {**chip_info, "usage": usage}
                hbm_value = int(usage.get("hbm", 0)) if usage else 0
        hbm_value = hbm_value * 1024  # 转换为MB
        collector = AscendCollector(npu_map, hbm_value)
    except Exception:  # noqa
        if all(v is None for v in info.values()):
            return None, None
    return info, collector


def get_version() -> str:
    result = subprocess.run(["npu-smi", "-v"], capture_output=True, check=True, text=True)
    return result.stdout.split(":")[-1].strip()


def get_cann_version() -> Optional[str]:
    """
    从 ascend_toolkit_install.info 文件中提取 Ascend-cann-toolkit 的版本号
    如果文件不存在或者格式不正确，返回 None
    """
    try:
        arch = platform.machine()
        if arch not in ("aarch64", "x86_64"):
            return None

        path = f"/usr/local/Ascend/ascend-toolkit/latest/{arch}-linux/ascend_toolkit_install.info"

        # 使用grep直接提取版本号，避免读取整个文件
        cmd = f"grep -m 1 '^version=' {path} | cut -d'=' -f2"
        proc = subprocess.run(cmd, shell=True, capture_output=True, text=True)

        return proc.stdout.strip() if proc.returncode == 0 else None

    except Exception:  # noqa
        return None


def map_npu() -> dict:
    """
    列出所有NPU设备，并包含芯片的映射关系
    """
    output = subprocess.run(["npu-smi", "info", "-m"], capture_output=True, check=True, text=True).stdout
    # npu_id -> chip_id -> {"id": chip_logic_id, "name": chip_name}
    npu_map = {}
    # output的第一行是表头，从第二行开始是数据，分别是NPU ID、芯片ID、芯片逻辑ID、芯片名称
    # 会包含Mcu芯片，这些芯片不参与计算，所以不存在逻辑ID，需要滤除
    lines = output.split("\n")[1:]
    for line in lines:
        line = line.split()
        if len(line) < 4:
            continue
        # 前三个值分别为NPU ID、芯片ID、芯片逻辑ID，最后可能会有多个值，是芯片名称
        npu_id, chip_id, chip_logic_id, *chip_name = line
        # 如果chip_logic_id不是数字，说明是Mcu芯片，不参与计算
        if not chip_logic_id.isdigit():
            continue
        chip_name = " ".join(chip_name)
        if npu_id not in npu_map:
            npu_map[npu_id] = {}
        npu_map[npu_id][chip_id] = {"id": chip_logic_id, "name": chip_name}
    return npu_map


def get_chip_usage(npu_id: str, chip_id: str):
    """
    获取某个NPU设备的芯片信息
    """
    output = subprocess.run(
        ["npu-smi", "info", "-t", "usages", "-i", npu_id, "-c", chip_id],
        capture_output=True,
        text=True,
    ).stdout
    usage = {}
    # 以空格分隔，找到HBM Capacity所在的行
    for line in output.split("\n"):
        if "hbm capacity" in line.lower():
            line = line.split(":")
            # HBM Capacity的值在最后一个
            hbm = line[-1].strip()
            if hbm.isdigit():
                usage["hbm"] = str(round(int(hbm) / 1024))  # 单位为GB
            break
    return usage


class AscendCollector(H):
    def __init__(self, npu_map, max_hbm_value: int):
        super().__init__()
        self.npu_map = npu_map
        self.max_hbm_value = max_hbm_value
        # NPU Utilization (%)
        self.util_key = generate_key("npu.{npu_index}.ptc")
        util_config = HardwareConfig(
            y_range=(0, 100),
            chart_index=random_index(),
            chart_name="NPU Utilization (%)",
        )
        self.per_util_configs = {}
        # NPU Memory Allocated (%)
        self.hbm_rate_key = generate_key("npu.{npu_index}.mem.ptc")
        hbm_rate_config = HardwareConfig(
            y_range=(0, 100),
            chart_index=random_index(),
            chart_name="NPU Memory Allocated (%)",
        )
        self.per_hbm_configs = {}
        # NPU Memory Allocated (MB)
        self.hbm_value_key = generate_key("npu.{npu_index}.mem.value")
        hbm_value_config = HardwareConfig(
            y_range=(0, max_hbm_value),
            chart_index=random_index(),
            chart_name="NPU Memory Allocated (MB)",
        )
        self.per_hbm_value_configs = {}

        # NPU Temperature (℃)
        self.temp_key = generate_key("npu.{npu_index}.temp")
        temp_config = HardwareConfig(
            y_range=(0, None),
            chart_index=random_index(),
            chart_name="NPU Temperature (℃)",
        )
        self.per_temp_configs = {}
        # NPU Power Usage (W)
        self.power_key = generate_key("npu.{npu_index}.power")
        power_config = HardwareConfig(
            y_range=(0, None),
            chart_index=random_index(),
            chart_name="NPU Power Usage (W)",
        )
        self.per_power_config = {}

        for npu_id in npu_map:
            for chip_id in npu_map[npu_id]:
                metric_name = f"NPU {npu_id}-{chip_id}"
                self.per_util_configs[metric_name] = util_config.clone(metric_name=metric_name)
                self.per_hbm_configs[metric_name] = hbm_rate_config.clone(metric_name=metric_name)
                self.per_hbm_value_configs[metric_name] = hbm_value_config.clone(metric_name=metric_name)
                self.per_temp_configs[metric_name] = temp_config.clone(metric_name=metric_name)
                self.per_power_config[metric_name] = power_config.clone(metric_name=metric_name)

    def collect(self) -> HardwareInfoList:
        result: HardwareInfoList = []
        for npu_id in self.npu_map:
            for chip_id in self.npu_map[npu_id]:
                result.extend(self.get_usage(npu_id, chip_id))
                result.append(self.get_chip_temp(npu_id, chip_id))
                result.append(self.get_chip_power(npu_id, chip_id))
        return result

    def get_usage(self, npu_id: str, chip_id: str) -> HardwareInfoList:
        """
        获取指定NPU设备的芯片HBM的用量信息和利用率
        """
        output = subprocess.run(
            ["npu-smi", "info", "-t", "usages", "-i", npu_id, "-c", chip_id],
            capture_output=True,
            text=True,
        ).stdout
        # 格式化获取NPU ID和芯片ID
        _id, metric_name = self.get_label(npu_id, chip_id)
        # 获取信息
        util_info = {
            "key": self.util_key.format(npu_index=_id),
            "name": f"{metric_name} Utilization (%)",
            "value": math.nan,
            "config": self.per_util_configs[metric_name],
        }
        hbm_info = {
            "key": self.hbm_rate_key.format(npu_index=_id),
            "name": f"{metric_name} Memory Allocated (%)",
            "value": math.nan,
            "config": self.per_hbm_configs[metric_name],
        }
        hbm_value_info = {
            "key": self.hbm_value_key.format(npu_index=_id),
            "name": f"{metric_name} Memory Allocated (MB)",
            "value": math.nan,
            "config": self.per_hbm_value_configs[metric_name],
        }
        for line in output.split("\n"):
            if "aicore usage rate" in line.lower():
                line = line.split(":")
                # 利用率的值在最后一个
                util = line[-1].strip()
                if util.isdigit():
                    util_info["value"] = float(util)
                continue

            if "hbm usage rate" in line.lower():
                line = line.split(":")
                # HBM Capacity的值在最后一个
                hbm = line[-1].strip()
                if hbm.isdigit():
                    hbm_info["value"] = float(hbm)
                    hbm_value_info["value"] = float(hbm) * self.max_hbm_value * 0.01
                continue
        return [util_info, hbm_info, hbm_value_info]

    def get_chip_temp(self, npu_id: str, chip_id: str) -> HardwareInfo:
        """
        获取芯片温度
        """
        output = subprocess.run(
            ["npu-smi", "info", "-t", "temp", "-i", npu_id, "-c", chip_id],
            capture_output=True,
            text=True,
        ).stdout.strip()
        temp = float(output.split(":")[-1].strip())
        _id, metric_name = self.get_label(npu_id, chip_id)
        return {
            "key": self.temp_key.format(npu_index=_id),
            "name": f"{metric_name} Temperature (℃)",
            "value": temp,
            "config": self.per_temp_configs[metric_name],
        }

    def get_chip_power(self, npu_id: str, chip_id: str) -> HardwareInfo:
        """
        获取芯片功耗信息
        """
        output = subprocess.run(
            ["npu-smi", "info", "-t", "power", "-i", npu_id, "-c", chip_id],
            capture_output=True,
            text=True,
        ).stdout.strip()
        power = float(output.split(":")[-1].strip())
        _id, metric_name = self.get_label(npu_id, chip_id)
        return {
            "key": self.power_key.format(npu_index=_id),
            "name": f"{metric_name} Power Usage (W)",
            "value": power,
            "config": self.per_power_config[metric_name],
        }

    @staticmethod
    def get_label(npu_id: str, chip_id: str):
        _id = f"{npu_id}-{chip_id}"
        return _id, f"NPU {_id}"



from mmengine.config import Config
from mmengine.registry import VISUALIZERS
import mmengine
import swanlab

print(f"MMEngine Version: {mmengine.__version__}")
print(f"SwanLab Version: {swanlab.__version__}")

cfg_text = """
# swanlab visualizer
custom_imports = dict(  # 引入SwanLab作为日志记录器，对于部分不支持custom_imports的项目可以直接初始化SwanlabVisBackend并加入vis_backends
    imports=["swanlab.integration.mmengine"], allow_failed_imports=False
)

vis_backends = [
    dict(
        type="SwanlabVisBackend",
        init_kwargs={ # swanlab.init 参数
            "project": "swanlab-mmengine",
            "experiment_name": "Your exp",  # 实验名称
            "description": "Note whatever you want",  # 实验的描述信息
        },
    ),
]

visualizer = dict(
    type="Visualizer",
    vis_backends=vis_backends,
    name="visualizer",
)

"""

cfg = Config.fromstring(cfg_text, ".py")

custom_vis = VISUALIZERS.build(cfg.visualizer)
print(custom_vis)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/6/12 21:17
@File: __init__.py
@IDE: pycharm
@Description:
    暴露子命令
"""
from .auth import login, logout
from .converter import convert
from .dashboard import watch
from .sync import sync



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/6/19 14:37
@File: formater.py
@IDE: pycharm
@Description:
    入参格式化器
"""
import json
import os
import re
from typing import Optional, Union, List

import yaml

from swanlab.toolkit import SwanKitCallback


def check_string(target: str) -> bool:
    """
    检查是否为字符串，且不能全空格，也不能为空字符串
    :param target: 待检查的字符串
    :return: bool
    :raises:
        :raise TypeError: name不是字符串
    """
    if not isinstance(target, str):
        raise TypeError(f"name: {target} is not a string: {type(target)}")
    # 利用正则表达式匹配非空格字符
    if re.match(r"^\s*$", target):
        return False
    # 利用正则表达式匹配非空字符串
    if re.match(r"^\s*$", target) or target == "":
        return False
    return True


def check_load_json_yaml(file_path: str, param_name):
    # 不是字符串
    if not isinstance(file_path, str):
        raise TypeError("{} must be a string, but got {}".format(param_name, type(file_path)))
    # 检查file_path的后缀是否是json/yaml，否则报错
    path_suffix = file_path.split(".")[-1]
    if not file_path.endswith((".json", ".yaml", ".yml")):
        raise ValueError(
            "{} must be a json or yaml file ('.json', '.yaml', '.yml'), "
            "but got {}, please check if the content of config_file is correct.".format(param_name, path_suffix)
        )
    # 转换为绝对路径
    file_path = os.path.abspath(file_path)
    # 读取配置文件
    # 如果文件不存在或者不是文件
    if (not os.path.exists(file_path)) or (not os.path.isfile(file_path)):
        raise FileNotFoundError("{} not found, please check if the file exists.".format(param_name))
    # 为空
    if os.path.getsize(file_path) == 0:
        raise ValueError("{} is empty, please check if the content of config_file is correct.".format(param_name))
    # 无权限读取
    if not os.access(file_path, os.R_OK):
        raise PermissionError("No permission to read {}, please check if you have the permission.".format(param_name))
    load = json.load if path_suffix == "json" else yaml.safe_load
    with open(file_path, "r", encoding='utf-8') as f:
        # 读取配置文件的内容
        file_data = load(f)
        # 如果读取的内容不是字典类型，则报错
        if not isinstance(file_data, dict):
            raise TypeError("The configuration file must be a dictionary, but got {}".format(type(file_data)))
    return file_data


# ---------------------------------- 实验、项目相关 ----------------------------------


def _auto_cut(name: str, value: str, max_len: int, cut: bool) -> str:
    """
    检查长度
    :param name: 参数名称
    :param value: 参数值
    :param max_len: 最大长度
    :return: str 检查后的字符串
    :raises
        :raise IndexError: cut为False且name超出长度
    """
    if len(value) > max_len:
        if cut:
            value = value[:max_len]
        else:
            raise IndexError(f"Name: {name} is too long, which must be less than {max_len} characters")
    return value


def check_proj_name_format(name: str, auto_cut: bool = True) -> str:
    """
    检查项目名称格式，最大长度为100个字符，支持 0-9, a-z, A-Z, _, -, +, .等字符

    Parameters
    ----------
    name : str
        待检查的字符串
    auto_cut : bool, optional
        如果超出长度，是否自动截断，默认为True
        如果为False，则超出长度会抛出异常

    Returns
    -------
    str
        检查后的字符串

    Raises
    ------
    TypeError
        name不是字符串，或者name为空字符串
    ValueError
        name不符合规定格式
    IndexError
        name超出长度
    """
    max_len = 100
    if not check_string(name) or not re.match(r"^[0-9a-zA-Z_\-+.]+$", name):
        raise ValueError(f"Project name `{name}` is invalid, which must be 0-9, a-z, A-Z, _ , -, +, .")
    name = name.strip()
    return _auto_cut("project", name, max_len, auto_cut)


def check_exp_name_format(name: str, auto_cut: bool = True) -> str:
    """
    检查实验名称格式，最大长度为250个字符，一个中文字符算一个字符
    其他不做限制，实验名称可以包含任何字符
    :param name: 实验名称
    :param auto_cut: 是否自动截断，默认为True
    :return: str 检查后的字符串
    """
    max_len = 250
    if not check_string(name):
        raise ValueError("Experiment name is an empty string")
    name = name.strip()
    return _auto_cut("experiment", name, max_len, auto_cut)


def check_desc_format(desc: str, auto_cut: bool = True) -> str:
    """
    检查描述格式，最大长度为255个字符，一个中文字符算一个字符
    :param desc: 描述信息
    :param auto_cut: 是否自动截断，默认为True
    :return: str 检查后的字符串
    """
    max_len = 255
    return _auto_cut("description", desc, max_len, auto_cut)


def check_tags_format(tags: List[str], auto_cut: bool = True) -> List[str]:
    """
    检查标签格式，最大长度为200个字符，一个中文字符算一个字符
    :param tags: 实验标签数列
    :param auto_cut: 是否自动截断，默认为True
    :return: str 检查后的字符串
    """
    max_len = 200
    new_tags = []
    for i in range(len(tags)):
        new_tags.append(_auto_cut(f"tags[{i}]", tags[i], max_len, auto_cut))
    return new_tags


def check_run_id_format(run_id: str = None) -> Optional[str]:
    """
    检查运行ID格式，要求：
    1. 只能包含小写字母和数字
    2. 长度为21个字符
    :param run_id: 运行ID字符串
    :return: str 检查后的字符串
    :raises ValueError: 如果运行ID不符合要求
    """
    if run_id is None:
        return None
    run_id_str = str(run_id)
    if not re.match(r"^[a-z0-9]{21}$", run_id_str):
        raise ValueError(f"id `{run_id}` is invalid, it must be 21 characters of lowercase letters and digits")
    return run_id_str


def check_key_format(key: str, auto_cut=True) -> str:
    """检查key字符串格式
    不能超过255个字符，可以包含任何字符，不允许.和/以及空格开头

    Parameters
    ----------
    key : str
        待检查的字符串
    auto_cut : bool, optional
        如果超出长度，是否自动截断，默认为True
        如果为False，则超出长度会抛出异常

    Returns
    -------
    str
        检查后的字符串

    Raises
    ------
    TypeError
        key不是字符串，或者key为空字符串
    ValueError
        key不符合规定格式
    IndexError
        key超出长度,此时auto_cut为False
    """
    max_len = 255
    if not isinstance(key, str):
        raise TypeError(f"tag: {key} is not a string")
    # 删除头尾空格
    key = key.lstrip().rstrip()
    if not check_string(key):
        raise ValueError(f"tag: {key} is an empty string")
    if key.startswith((".", "/")):
        raise ValueError(f"tag: {key} can't start with '.' or '/' and blank space")
    if key.endswith((".", "/")):  # cannot create folder end with '.' or '/'
        raise ValueError(f"tag: {key} can't end with '.' or '/' and blank space")
    # 检查长度
    return _auto_cut("tag", key, max_len, auto_cut)


def check_callback_format(callback: Optional[Union[SwanKitCallback, List[SwanKitCallback]]]) -> List[SwanKitCallback]:
    """
    检查回调函数格式
    :param callback: 回调函数对象
    :return: List[SwanKitCallback] 回调函数列表
    """
    if callback is None:
        return []
    if isinstance(callback, SwanKitCallback):
        return [callback]
    if not isinstance(callback, list):
        raise TypeError(f"Only support SwanKitCallback or List[SwanKitCallback], but got {type(callback)}")
    return callback



import torch
from lightning import LightningModule
from torch.utils.data import Dataset


class RandomDataset(Dataset):
    def __init__(self, size, num_samples):
        self.len = num_samples
        self.data = torch.randn(num_samples, size)

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.len


class BoringModel(LightningModule):
    def __init__(self):
        super().__init__()
        self.layer = torch.nn.Linear(32, 2)
        self.training_step_outputs = []
        self.validation_step_outputs = []
        self.test_step_outputs = []

    def forward(self, x):
        return self.layer(x)

    def loss(self, batch, prediction):
        # An arbitrary loss to have a loss that updates the model weights during `Trainer.fit` calls
        return torch.nn.functional.mse_loss(prediction, torch.ones_like(prediction))

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(self.layer.parameters(), lr=0.1)
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1)
        return [optimizer], [lr_scheduler]

    def training_step(self, batch, _):
        output = self.layer(batch)
        loss = self.loss(batch, output)
        self.log("loss", loss)
        self.training_step_outputs.append(loss)
        return loss

    def on_train_epoch_end(self):
        _ = torch.stack(self.training_step_outputs).mean()
        self.training_step_outputs.clear()  # free memory

    def validation_step(self, batch, _):
        output = self.layer(batch)
        loss = self.loss(batch, output)
        self.validation_step_outputs.append(loss)
        return loss

    def on_validation_epoch_end(self) -> None:
        _ = torch.stack(self.validation_step_outputs).mean()
        self.validation_step_outputs.clear()  # free memory

    def test_step(self, batch, _):
        output = self.layer(batch)
        loss = self.loss(batch, output)
        self.log("fake_test_acc", loss)
        self.test_step_outputs.append(loss)
        return loss

    def on_test_epoch_end(self) -> None:
        _ = torch.stack(self.test_step_outputs).mean()
        self.test_step_outputs.clear()  # free memory



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/26 16:03
@File: pytest_main.py
@IDE: pycharm
@Description:
    测试SwanLabRun主类
"""
import io
import json
import math
import os

import nanoid
import numpy as np
import pytest
import soundfile as sf
from PIL import Image as PILImage
from nanoid import generate

import swanlab
import tutils as T
from swanlab import Image, Audio, Text
from swanlab.data.modules import Line
from swanlab.data.run.main import SwanLabRun, get_run, SwanLabRunState, swanlog, get_url, get_project_url
from swanlab.env import SwanLabEnv
from tutils import TEMP_PATH
from tutils.setup import UseMockRunState


@pytest.fixture(scope="function", autouse=True)
def setup_function():
    """
    在当前测试文件下的每个测试函数执行前后执行
    """
    swanlog.disable_log()
    yield
    swanlog.enable_log()
    if get_run() is not None:
        get_run().finish()


class TestSwanLabRunInit:

    def test_before_init(self):
        run = get_run()
        assert run is None
        assert SwanLabRun.get_state() == SwanLabRunState.NOT_STARTED

    def test_after_init(self):
        os.environ[SwanLabEnv.MODE.value] = "disabled"
        with UseMockRunState():
            run = SwanLabRun()
            assert swanlog.proxied is False
            assert run is not None
            assert get_run().__str__() == run.__str__()
            _run = run.finish()
            assert get_run() is None
            assert _run.__str__() == run.__str__()
            assert SwanLabRunState.SUCCESS == _run.state

    def test_duplicate_init(self):
        os.environ[SwanLabEnv.MODE.value] = "disabled"
        with UseMockRunState():
            run = SwanLabRun()
            with pytest.raises(RuntimeError) as e:
                SwanLabRun()
            assert swanlog.proxied is False
            assert str(e.value) == "SwanLabRun has been initialized"
            assert run.__str__() == get_run().__str__()
            _run = run.finish()
            assert swanlog.proxied is False


class TestSwanLabRunState:
    """
    测试SwanLabRun的状态变化
    """

    @staticmethod
    def setup_method():
        os.environ[SwanLabEnv.MODE.value] = "disabled"

    def test_not_started(self):
        assert SwanLabRun.get_state() == SwanLabRunState.NOT_STARTED

    def test_running(self):
        with UseMockRunState():
            run = SwanLabRun()
            assert run.state == SwanLabRunState.RUNNING
            assert run.running is True

    def test_crashed(self):
        with UseMockRunState():
            run = SwanLabRun()
            run.finish(SwanLabRunState.CRASHED, error="error")
            assert run.state == SwanLabRunState.CRASHED
            assert run.crashed is True

    def test_success(self):
        with UseMockRunState():
            run = SwanLabRun()
            run.finish(SwanLabRunState.SUCCESS)
            assert run.state == SwanLabRunState.SUCCESS
            assert run.success is True


class TestSwanLabRunLog:
    """
    测试SwanLabRun的日志解析功能，不包含操作员，由于情况比较多，这里分情况测试
    1. 一开始正常输入，然后重复输入，然后输入新的key的重复的step
    2. 一开始输入错误的类型，然后正常输入
    3. 一开始输入正确的类型，然后输入错误的类型
    """

    @staticmethod
    def setup_method():
        os.environ[SwanLabEnv.MODE.value] = "disabled"

    # ---------------------------------- 返回的列信息将包含id ----------------------------------

    def test_log_column_with_id(self):
        """
        id为当前创建列前历史列数量
        """
        with UseMockRunState() as run_state:
            run = SwanLabRun()
            data = {"a": 1, "b": 2}
            ll = run.log(data)
            assert len(ll) == 2
            assert ll["a"].column_info.kid == "0"
            assert ll["b"].column_info.kid == "1"

    # ---------------------------------- 解析log数字/Line ----------------------------------
    def test_log_number_ok(self):
        with UseMockRunState():
            run = SwanLabRun()
            data = {"a": 1, "b": 0.1, "c": {"d": 2}, "math.nan": math.nan, "math.inf": math.inf}
            ll = run.log(data)
            assert len(ll) == 5
            # 都没有错误
            assert all([ll[k].is_error is False for k in ll])
            assert ll["a"].data == 1
            assert ll["b"].data == 0.1
            assert ll["c.d"].data == 2
            assert ll["math.nan"].data == Line.nan
            assert ll["math.inf"].data == Line.inf
            assert all([ll[k].column_info.chart_type == ll[k].column_info.chart_type.LINE for k in ll])
            # 没有其他多余的内容
            assert all([ll[k].buffers is None for k in ll])
            assert all([ll[k].data is not None for k in ll])
            # 没有指定step的话从0开始
            assert all([ll[k].metric_step == 0 for k in ll])
            ll2 = run.log(data, step=3)
            assert all(ll2[k].metric_step == 3 for k in ll2)
            # 重复的step会被忽略
            ll3 = run.log(data, step=3)
            assert all(ll3[k].is_error for k in ll3)
            assert all(ll3[k].error.duplicated for k in ll3)
            assert all(ll3[k].column_error is None for k in ll3)
            assert all(ll3[k].error is not None for k in ll3)
            # 如果是新的key的重复step，会被添加
            ll4 = run.log({"tmp": 1}, step=3)
            assert all(ll4[k].is_error is False for k in ll4)
            assert all(ll4[k].metric_step == 3 for k in ll4)

    def test_log_number_use_line(self):
        """
        使用Line类型log，本质上应该与数字类型一样，数字类型是Line类型的语法糖
        """
        with UseMockRunState():
            run = SwanLabRun()
            data = {
                "a": Line(1),
                "b": Line(0.1),
                "c": {"d": Line(2)},
                "math.nan": Line(math.nan),
                "math.inf": Line(math.inf),
            }
            ll = run.log(data)
            assert len(ll) == 5
            # line(1)和[line(1)]是一样的
            ll2 = run.log({"a": [Line(1)]})
            assert ll2["a"].data == ll["a"].data
            # Line类型不允许多个元素
            ll3 = run.log({"a": [Line(1), Line(2)]})
            assert ll3["a"].is_error is True
            assert ll3["a"].column_error is None
            assert ll3["a"].error.expected == 'float'

    def test_log_number_error_type(self):
        """
        使用错误的，不受支持的类型log
        """
        with UseMockRunState():
            run = SwanLabRun()
            data = {"a": "a", "b": [1], "c": 1}
            ll = run.log(data)
            assert len(ll) == 3
            # 前两个错误，最后一个正确
            assert ll["a"].is_error is True
            assert ll["b"].is_error is True
            assert ll["c"].is_error is False

            assert ll["a"].column_error is not None
            assert ll["b"].column_error is not None
            assert ll["c"].column_error is None

            assert ll["a"].column_error.expected == 'float'
            assert ll["b"].column_error.expected == 'float'
            assert ll["c"].data == 1

            assert ll["a"].error == ll['a'].column_error
            assert ll["b"].error == ll['b'].column_error

    # ---------------------------------- 解析log文字 ----------------------------------

    def test_log_text_ok(self):
        with UseMockRunState():
            run = SwanLabRun()
            data = {"a": Text("abc"), "b": Text("def")}
            ll = run.log(data)
            assert len(ll) == 2
            assert all([ll[k].is_error is False for k in ll])
            assert ll["a"].data == ["abc"]
            assert ll["b"].data == ["def"]
            assert all([ll[k].column_info.chart_type == ll[k].column_info.chart_type.TEXT for k in ll])
            assert all([ll[k].buffers is None for k in ll])
            assert all([ll[k].data is not None for k in ll])
            assert all([ll[k].metric_step == 0 for k in ll])
            ll2 = run.log(data, step=3)
            assert all(ll2[k].metric_step == 3 for k in ll2)
            # list
            ll3 = run.log({"a": [Text("abc"), Text("def")]})
            assert ll3["a"].data == ["abc", "def"]
            data = {"a": [Text("abc")] * 109}
            ll4 = run.log(data, step=4)
            assert ll4["a"].data == ["abc"] * 108

    # ---------------------------------- 解析log Audio ----------------------------------

    @staticmethod
    def check_audio(ll: dict, key: str):
        with UseMockRunState():
            assert len(ll[key].buffers) == 1
            buffer = ll[key].buffers[0]
            audio, _samplerate = sf.read(io.BytesIO(buffer.getvalue()))
            assert _samplerate == 5000
            # FIXME 一维的？
            assert len(audio.shape) == 1

    def test_log_audio_path(self):
        """
        通过路径添加音频
        """
        with UseMockRunState():
            run = SwanLabRun()
            # 正确的音频路径
            path = os.path.join(TEMP_PATH, "test.wav")
            samplerate = 5000
            sf.write(path, np.random.rand(1000), samplerate)
            data = {"a": Audio(path)}
            ll = run.log(data)
            self.check_audio(ll, "a")

    def test_log_audio_numpy(self):
        """
        通过numpy数组添加音频
        """
        with UseMockRunState():
            run = SwanLabRun()
            samplerate = 5000
            data = {"a": Audio(np.random.rand(1000), samplerate)}
            ll = run.log(data)
            self.check_audio(ll, "a")

    def test_log_audio_error(self):
        """
        错误的情况
        """
        with UseMockRunState():
            run = SwanLabRun()
            # 错误的路径
            with pytest.raises(ValueError):
                run.log({"a": Audio("error.wav")})
            # 错误的类型
            with pytest.raises(TypeError):
                run.log({"a": Audio(1)})  # noqa
            # 错误的矩阵编码类型
            with pytest.raises(TypeError):
                run.log({"a": Audio(np.random.rand(1000).astype(np.int8))})
            # 错误的矩阵维数
            with pytest.raises(TypeError):
                run.log({"a": Audio(np.random.rand(1000, 5))})

    # ---------------------------------- 图像 ----------------------------------
    @staticmethod
    def check_image(ll: dict, key: str):
        assert len(ll[key].buffers) == 1
        buffer = ll[key].buffers[0]
        image = PILImage.open(io.BytesIO(buffer.getvalue()))
        assert image.size == (100, 100)

    def test_log_image_path(self):
        """
        通过路径添加图像
        """
        with UseMockRunState():
            run = SwanLabRun()
            # 正确的图像路径
            path = os.path.join(TEMP_PATH, "test.png")
            PILImage.new("RGB", (100, 100)).save(path)
            data = {"a": Image(path)}
            ll = run.log(data)
            self.check_image(ll, "a")

    def test_log_image_numpy(self):
        """
        通过numpy数组添加图像
        """
        with UseMockRunState():
            run = SwanLabRun()
            data = {"a": Image(np.random.rand(100, 100, 3))}
            ll = run.log(data)
            self.check_image(ll, "a")

    # ---------------------------------- echarts ----------------------------------
    def test_log_echarts(self):
        with UseMockRunState():
            run = SwanLabRun()
            data = {"a": swanlab.echarts.Bar().add_xaxis(["X", "Y"]).add_yaxis("数值", [10, 20])}
            ll = run.log(data)
            assert len(ll['a'].buffers) == 1
            buffer = ll['a'].buffers[0]
            chart_data = json.loads(buffer.getvalue().decode("utf-8"))
            assert isinstance(chart_data, dict)

    # 其他类似...


class TestGetUrl:

    @pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
    def test_ok(self):
        """
        初始化一个实验
        """
        exp_name = generate()
        project_name = generate()
        swanlab.init(project=project_name, experiment_name=exp_name, mode="cloud")
        assert get_url() is not None and get_url() != ""

    def test_local(self):
        """
        本地模式
        """
        swanlab.init(mode="local")
        assert get_url() is None

    def test_before_init(self):
        """
        未初始化
        """
        assert get_url() is None


class TestGetProjectUrl:

    @pytest.mark.skipif(T.is_skip_cloud_test, reason="skip cloud test")
    def test_ok(self):
        """
        初始化一个实验
        """
        exp_name = generate()
        project_name = generate()
        swanlab.init(project=project_name, experiment_name=exp_name, mode="cloud")
        assert get_project_url() is not None and get_project_url() != ""
        assert get_project_url().endswith(project_name)

    def test_local(self):
        """
        本地模式
        """
        swanlab.init(mode="local")
        assert get_project_url() is None

    def test_before_init(self):
        """
        未初始化
        """
        assert get_project_url() is None


@pytest.mark.parametrize(
    "mode,run_id",
    [
        ("cloud", nanoid.generate('0123456789abcdefghijklmnopqrstuvwxyz', 21)),
        ("local", None),
        ("disabled", None),
        ("offline", None),
    ],
)
def test_run_id(mode, run_id):
    """
    测试获取当前运行的ID
    只有 cloud 模式能获取到运行ID，其他模式返回None
    """
    os.environ[SwanLabEnv.MODE.value] = mode
    with UseMockRunState() as state:
        state.store.run_id = run_id if mode == "cloud" else state.store.run_id
        run = SwanLabRun()
        assert run.id == run_id


def test_run_id_none():
    """
    测试在未初始化时获取运行ID
    """
    with UseMockRunState() as state:
        state.store.run_id = None
        with pytest.raises(AssertionError):
            SwanLabRun()



"""
@author: caddiesnew
@file: hygon.py
@time: 2025-06-04 12:08:51
@description: Hygon DCU 信息采集
"""

import json
import math
import platform
import subprocess
from typing import Optional, Tuple

from ..type import HardwareCollector as H
from ..type import HardwareConfig, HardwareFuncResult, HardwareInfoList
from ..utils import generate_key, random_index


def get_hygon_dcu_info() -> HardwareFuncResult:
    """
    获取 Hygon DCU信息，包括驱动版本、设备信息等
    """
    # Hygon DCU默认为Linux系统
    if platform.system() != "Linux":
        return None, None

    info = {"driver": None, "dcu": None}
    collector = None
    try:
        driver, dcu_map = map_hygon_dcu()
        info["driver"] = driver
        info["dcu"] = dcu_map
        max_mem_value = 0
        for dcu_id in dcu_map:
            mem_value = int(dcu_map[dcu_id]["memory"][:-2])
            max_mem_value = max(max_mem_value, mem_value)
        max_mem_value *= 1024
        collector = DCUCollector(dcu_map, max_mem_value)
    except Exception:  # noqa
        if all(v is None for v in info.values()):
            return None, None
    return info, collector


def map_hygon_dcu() -> Tuple[Optional[str], dict]:
    """
    获取 Hygon DCU 信息，包括驱动版本、设备信息等，例如：
    Driver Version: 6.2.28
    {'0': {'name': 'K500SM_AI', 'memory': '64GB'}, '1': {'name': 'K500SM_AI', 'memory': '64GB',...}}
    """
    driver_info = subprocess.run(["hy-smi", "--showdriverversion"], capture_output=True, check=True, text=True).stdout
    driver_version = None
    for line in driver_info.split("\n"):
        if "Driver Version" in line:
            driver_version = line.split(":")[-1]
            break

    dcu_product_map_str = subprocess.run(
        ["hy-smi", "--showproductname", "--json"],
        capture_output=True,
        check=True,
        text=True,
    ).stdout
    dcu_product_map = json.loads(dcu_product_map_str)

    dcu_available_mem_str = subprocess.run(
        ["hy-smi", "--showmemavailable", "--json"],
        capture_output=True,
        check=True,
        text=True,
    ).stdout
    dcu_available_mem_map = json.loads(dcu_available_mem_str)

    dcu_map = {}
    for idx, device_id in enumerate(dcu_product_map.keys()):
        product_info = dcu_product_map.get(device_id, {})
        device_name = product_info.get("Card Series", "Unknown")

        available_info = dcu_available_mem_map.get(device_id, {})
        available_mem = available_info.get("Available memory size (MiB)", "0")

        try:
            total_mem_gb = round(int(available_mem) / 1024)
            mem_str = f"{total_mem_gb}GB"
        except (ValueError, TypeError):
            mem_str = "0GB"

        dcu_map[str(idx)] = {"name": device_name, "memory": mem_str}
    return driver_version, dcu_map


class DCUCollector(H):
    def __init__(self, dcu_map, max_mem_value):
        super().__init__()
        self.dcu_map = dcu_map
        self.max_mem_value = max_mem_value

        # DCU Utilization (%)
        self.util_key = generate_key("dcu.{dcu_index}.pct")
        util_config = HardwareConfig(
            y_range=(0, 100),
            chart_index=random_index(),
            chart_name="DCU Utilization (%)",
        )
        self.per_util_configs = {}

        # DCU Memory Allocated (%)
        self.memory_key = generate_key("dcu.{dcu_index}.mem.pct")
        memory_config = HardwareConfig(
            y_range=(0, 100),
            chart_index=random_index(),
            chart_name="DCU Memory Allocated (%)",
        )
        self.per_memory_configs = {}

        # DCU Memory Allocated (MB)
        self.mem_value_key = generate_key("dcu.{dcu_index}.mem.value")
        mem_value_config = HardwareConfig(
            y_range=(0, self.max_mem_value),
            chart_index=random_index(),
            chart_name="DCU Memory Allocated (MB)",
        )
        self.per_mem_value_configs = {}

        # DCU Temperature (°C)
        self.temp_key = generate_key("dcu.{dcu_index}.temp")
        temp_config = HardwareConfig(
            y_range=(0, None),
            chart_index=random_index(),
            chart_name="DCU Temperature (°C)",
        )
        self.per_temp_configs = {}

        # DCU Power (W)
        self.power_key = generate_key("dcu.{dcu_index}.power")
        power_config = HardwareConfig(
            y_range=(0, None),
            chart_index=random_index(),
            chart_name="DCU Power (W)",
        )
        self.per_power_configs = {}

        for dcu_id in self.dcu_map:
            metric_name = f"DCU {dcu_id}"
            self.per_util_configs[metric_name] = util_config.clone(metric_name=metric_name)
            self.per_memory_configs[metric_name] = memory_config.clone(metric_name=metric_name)
            self.per_mem_value_configs[metric_name] = mem_value_config.clone(metric_name=metric_name)
            self.per_temp_configs[metric_name] = temp_config.clone(metric_name=metric_name)
            self.per_power_configs[metric_name] = power_config.clone(metric_name=metric_name)

    def collect(self) -> HardwareInfoList:
        result: HardwareInfoList = []
        usage_methods = [
            self.get_utilization_usage,
            self.get_memory_usage,
            self.get_mem_value_usage,
            self.get_temperature_usage,
            self.get_power_usage,
        ]

        for method in usage_methods:
            result.extend(method().values())
        return result

    def get_utilization_usage(self) -> dict:
        """
        获取指定DCU设备的利用率
        """
        output_str = subprocess.run(
            ["hy-smi", "--showuse", "--json"],
            capture_output=True,
            text=True,
        ).stdout
        output_json = json.loads(output_str)
        util_infos = {}
        for idx, (dcu_key, dcu_info) in enumerate(output_json.items()):
            dcu_id = str(idx)

            util_infos[dcu_id] = {
                "key": self.util_key.format(dcu_index=dcu_id),
                "name": f"DCU {dcu_id} Utilization (%)",
                "value": math.nan,
                "config": self.per_util_configs[f"DCU {dcu_id}"],
            }
            dcu_utilization = dcu_info["DCU use (%)"]
            util_infos[dcu_id]["value"] = float(dcu_utilization)
        return util_infos

    def get_memory_usage(self) -> dict:
        """
        获取指定DCU设备的内存占用率
        """
        output_str = subprocess.run(
            ["hy-smi", "--showmemuse", "--json"],
            capture_output=True,
            text=True,
        ).stdout
        output_json = json.loads(output_str)
        mem_infos = {}

        for idx, (dcu_key, dcu_info) in enumerate(output_json.items()):
            dcu_id = str(idx)
            mem_infos[dcu_id] = {
                "key": self.memory_key.format(dcu_index=dcu_id),
                "name": f"DCU {dcu_id} Memory Allocated (%)",
                "value": math.nan,
                "config": self.per_memory_configs[f"DCU {dcu_id}"],
            }

            dcu_mem_use = dcu_info["DCU memory use (%)"]
            mem_infos[dcu_id]["value"] = float(dcu_mem_use)
        return mem_infos

    def get_mem_value_usage(self) -> dict:
        """
        获取指定DCU设备的内存使用量（MB）
        """
        output_str = subprocess.run(
            ["hy-smi", "--showmemuse", "--json"],
            capture_output=True,
            text=True,
        ).stdout
        output_json = json.loads(output_str)
        mem_value_infos = {}

        for idx, (dcu_key, dcu_info) in enumerate(output_json.items()):
            dcu_id = str(idx)
            mem_value_infos[dcu_id] = {
                "key": self.mem_value_key.format(dcu_index=dcu_id),
                "name": f"DCU {dcu_id} Memory Allocated (MB)",
                "value": math.nan,
                "config": self.per_mem_value_configs[f"DCU {dcu_id}"],
            }

            dcu_mem_use_rate = dcu_info["DCU memory use (%)"]
            mem_value_infos[dcu_id]["value"] = float(dcu_mem_use_rate) * 0.01 * self.max_mem_value
        return mem_value_infos

    def get_temperature_usage(self) -> dict:
        """
        获取指定DCU设备的温度(°C) (采集 Junction 核心温度)
        {
          "card0": {
            "Temperature (Sensor edge) (C)": "52.0",
            "Temperature (Sensor junction) (C)": "51.0", // collected
            "Temperature (Sensor mem) (C)": "52.0"
          }
        }
        """
        output_str = subprocess.run(
            ["hy-smi", "--showtemp", "--json"],
            capture_output=True,
            text=True,
        ).stdout
        output_json = json.loads(output_str)
        temp_infos = {}
        for idx, (dcu_key, dcu_info) in enumerate(output_json.items()):
            dcu_id = str(idx)

            temp_infos[dcu_id] = {
                "key": self.temp_key.format(dcu_index=dcu_id),
                "name": f"DCU {dcu_id} Temperature (°C)",
                "value": math.nan,
                "config": self.per_temp_configs[f"DCU {dcu_id}"],
            }
            dcu_temp = dcu_info["Temperature (Sensor junction) (C)"]
            temp_infos[dcu_id]["value"] = float(dcu_temp)
        return temp_infos

    def get_power_usage(self) -> dict:
        """
        获取指定DCUhy-smi --设备的功耗(W)
        """
        output_str = subprocess.run(
            ["hy-smi", "--showpower", "--json"],
            capture_output=True,
            text=True,
        ).stdout
        output_json = json.loads(output_str)
        power_infos = {}
        for idx, (dcu_key, dcu_info) in enumerate(output_json.items()):
            dcu_id = str(idx)

            power_infos[dcu_id] = {
                "key": self.power_key.format(dcu_index=dcu_id),
                "name": f"DCU {dcu_id} Power (W)",
                "value": math.nan,
                "config": self.per_power_configs[f"DCU {dcu_id}"],
            }
            dcu_power = dcu_info["Average Graphics Package Power (W)"]
            power_infos[dcu_id]["value"] = float(dcu_power)
        return power_infos



"""
@author: cunyue
@file: __init__.py
@time: 2025/6/16 13:29
@description: swanlab 客户端，负责发送 http 请求
"""

import json
from datetime import datetime
from typing import Optional, Tuple, Dict, Union, List, AnyStr

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from swanlab.error import NetworkError, ApiError
from swanlab.log import swanlog
from swanlab.package import get_package_version
from swanlab.toolkit import MediaBuffer
from .cos import CosClient
from .model import ProjectInfo, ExperimentInfo
from .. import auth


def decode_response(resp: requests.Response) -> Union[Dict, AnyStr, List]:
    """
    解码响应，返回信息
    低版本requests库没有JSONDecodeError，所以需要捕获两种异常
    """
    try:
        return resp.json()
    except json.decoder.JSONDecodeError:
        return resp.text
    except requests.JSONDecodeError:
        return resp.text


class Client:
    """
    封装请求函数，添加get、post、put、delete方法
    """

    REFRESH_TIME = 60 * 60 * 24 * 7  # 7天
    """
    刷新时间，单位秒，如果sid过期时间减去当前时间小于这个时间，就刷新sid
    """

    def __init__(self, login_info: auth.LoginInfo):
        """
        初始化会话
        """
        self.__login_info = login_info
        # 当前cos信息
        self.__cos: Optional[CosClient] = None
        # 当前项目信息
        self.__proj: Optional[ProjectInfo] = None
        # 当前实验信息
        self.__exp: Optional[ExperimentInfo] = None
        # 当前进程会话
        self.__session: Optional[requests.Session] = None
        # 当前项目所属的username
        self.__groupname = login_info.username
        self.__version = get_package_version()
        # 创建会话
        self.__create_session()
        # 标识当前实验会话（flagId）是否被其他进程顶掉
        self.pending = False

    # ---------------------------------- 一些辅助属性 ----------------------------------
    @property
    def exp(self) -> ExperimentInfo:
        assert self.__exp is not None, "Experiment not mounted, please call mount_exp() first"
        return self.__exp

    @property
    def proj(self) -> ProjectInfo:
        assert self.__proj is not None, "Project not mounted, please call mount_project() first"
        return self.__proj

    @property
    def base_url(self):
        return self.__login_info.api_host

    @property
    def api_host(self):
        return self.__login_info.web_host

    @property
    def api_key(self):
        return self.__login_info.api_key

    @property
    def web_host(self):
        return self.__login_info.web_host

    @property
    def groupname(self):
        """
        当前项目所属组名
        """
        return self.__groupname

    @property
    def username(self):
        """
        当前登录的用户名
        """
        return self.__login_info.username

    @property
    def cos(self):
        return self.__cos

    @property
    def projname(self):
        return self.__proj.name

    @property
    def history_exp_count(self):
        return self.__proj.history_exp_count

    @property
    def exp_id(self):
        return self.__exp.cuid

    @property
    def expname(self):
        return self.__exp.name

    @property
    def sid_expired_at(self):
        """
        获取sid的过期时间，字符串格式转时间
        """
        return datetime.strptime(self.__login_info.expired_at, "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def web_proj_url(self):
        return f"{self.web_host}/@{self.groupname}/{self.projname}"

    @property
    def web_exp_url(self):
        return f"{self.web_proj_url}/runs/{self.exp_id}"

    # ---------------------------------- http方法 ----------------------------------

    def __before_request(self):
        """
        请求前的钩子
        """
        # FIXME datetime.utcnow() -> datetime.now(datetime.UTC)
        if (self.sid_expired_at - datetime.utcnow()).total_seconds() <= self.REFRESH_TIME:
            # 刷新sid，新建一个会话
            swanlog.debug("Refresh sid...")
            self.__login_info = auth.login_by_key(self.__login_info.api_key, save=False)
            self.__session.headers["cookie"] = f"sid={self.__login_info.sid}"

        # 携带实验会话Id
        if self.__exp is not None:
            self.__session.headers["flagId"] = self.__exp.flag_id

    def __create_session(self):
        """
        创建会话，这将在HTTP类实例化时调用
        添加了重试策略
        """
        session = requests.Session()
        retry = Retry(
            total=3,
            backoff_factor=0.1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE", "PATCH"]),
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("https://", adapter)

        session.headers["swanlab-sdk"] = self.__version
        session.cookies.update({"sid": self.__login_info.sid})

        # 注册响应钩子
        def response_interceptor(response: requests.Response, *args, **kwargs):
            """
            捕获所有的http不为2xx的错误，以ApiError的形式抛出
            """
            if response.status_code // 100 != 2:
                traceid = f"Trace id: {response.headers.get('traceid')}"
                request = f"{response.request.method.upper()} {response.url}"
                resp = f"{response.status_code} {response.reason}"
                raise ApiError(response, traceid, request, resp)

        session.hooks["response"] = response_interceptor

        self.__session = session

    def post(self, url: str, data: Union[dict, list] = None):
        """
        post请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.post(url, json=data)
        return decode_response(resp), resp

    def put(self, url: str, data: dict = None):
        """
        put请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.put(url, json=data)
        return decode_response(resp), resp

    def get(self, url: str, params: dict = None):
        """
        get请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.get(url, params=params)
        return decode_response(resp), resp

    def patch(self, url: str, data: dict = None):
        """
        patch请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.patch(url, json=data)
        return decode_response(resp), resp

    # ---------------------------------- 对象存储相关方法 ----------------------------------

    def __get_cos(self):
        self.__cos = CosClient(
            data=self.get(f"/project/{self.groupname}/{self.projname}/runs/{self.exp_id}/sts")[0],
        )

    def upload(self, buffer: MediaBuffer):
        """
        上传文件，需要注意的是file_path应该为unix风格而不是windows风格
        :param buffer: 自定义文件内存对象
        """
        if self.__cos.should_refresh:
            self.__get_cos()
        return self.__cos.upload(buffer)

    def upload_files(self, buffers: List[MediaBuffer]) -> Dict[str, Union[bool, List]]:
        """
        批量上传文件，keys和local_paths的长度应该相等
        :param buffers: 文件内存对象
        :return: 返回上传结果, 包含success_all和detail两个字段，detail为每一个文件的上传结果（通过index索引对应）
        """
        if self.__cos.should_refresh:
            swanlog.debug("Refresh cos...")
            self.__get_cos()
        return self.__cos.upload_files(buffers)

    # ---------------------------------- 接入后端api ----------------------------------

    def mount_project(self, name: str, username: str = None, public: bool = None):
        """
        创建项目，如果项目已存在，则获取项目信息
        :param name: 项目名称
        :param username: 项目所属的用户名
        :param public: 项目是否公开
        :return: 项目信息
        """
        try:
            data = {"name": name}
            if username is not None:
                data["username"] = username
            if public is not None:
                data["visibility"] = "PUBLIC" if public else "PRIVATE"
            resp_data, _ = self.post(f"/project", data=data)
        except ApiError as e:
            if e.resp.status_code == 409:
                # 项目已经存在，从对象中解析信息
                resp_data = decode_response(e.resp)
            elif e.resp.status_code == 404 and e.resp.reason == "Not Found":
                # WARNING: 早期 （私有化） swanlab 后端没有 /project 接口，需要使用 /project/{username} 接口，此时没有默认空间的特性
                self.__groupname = self.__groupname if username is None else username
                try:
                    visibility = "PUBLIC" if public else "PRIVATE"
                    resp_data, _ = self.post(
                        f"/project/{self.groupname}", data={"name": name, "visibility": visibility}
                    )
                except ApiError as e:
                    # 如果为409，表示已经存在，获取项目信息
                    if e.resp.status_code == 409:
                        resp_data, _ = self.get(f"/project/{self.groupname}/{name}")
                    elif e.resp.status_code == 404:
                        # 组织/用户不存在
                        raise ValueError(f"Space `{self.groupname}` not found")
                    elif e.resp.status_code == 403:
                        # 权限不足
                        raise ValueError(f"Space permission denied: " + self.groupname)
                    else:
                        raise e
            else:
                # 此接口为后端处理，sdk 在理论上不会出现其他错误，因此不需要处理其他错误
                raise e
        # 设置当前项目所属的用户名
        self.__groupname = resp_data['username']
        # 获取详细信息
        resp_data_info, _ = self.get(f"/project/{self.groupname}/{name}")
        self.__proj = ProjectInfo(resp_data_info)

    def mount_exp(
        self,
        exp_name,
        colors: Tuple[str, str],
        description: str = None,
        tags: List[str] = None,
        created_at: str = None,
        cuid: str = None,
        must_exist: bool = False,
    ) -> bool:
        """
        初始化实验，获取存储信息
        :param exp_name: 所属实验名称
        :param colors: 实验颜色，有两个颜色
        :param description: 实验描述
        :param tags: 实验标签
        :param created_at: 实验创建时间，格式为 ISO 8601
        :param cuid: 实验的唯一标识符，如果不提供则由后端生成
        :param must_exist: 如果 cuid 被传递，是否限制实验必须存在

        :raises RuntimeError: 如果实验不存在且must_exist为True
        :raises NotImplementedError: 如果项目未挂载

        :return: 返回实验为新建的还是更新的，为 True 时为新建实验
        """
        if self.__proj is None:
            raise NotImplementedError("Project not mounted, please call mount_project() first")
        if must_exist:
            assert cuid is not None, "cuid must be provided when must_exist is True"
            try:
                self.get(f"/project/{self.groupname}/{self.__proj.name}/runs/{cuid}")
            except ApiError as e:
                if e.resp.status_code == 404 and e.resp.reason == "Not Found":
                    raise RuntimeError(f"Experiment {cuid} does not exist in project {self.projname}")

        labels = [{"name": tag} for tag in tags] if tags else []
        post_data = {
            "name": exp_name,
            "description": description,
            "createdAt": created_at,
            "colors": list(colors),
            "labels": labels if len(labels) else None,
            "cuid": cuid,
        }
        post_data = {k: v for k, v in post_data.items() if v is not None}  # 移除值为None的键

        # 这部分错误将不会被上层捕获，直接抛出异常
        try:
            data, resp = self.post(f"/project/{self.groupname}/{self.__proj.name}/experiment", post_data)
        except ApiError as e:
            if e.resp.status_code == 400 and e.resp.reason == "Bad Request":
                # 指定的 cuid 对应的实验是克隆实验
                raise ValueError(
                    f"Experiment with CUID {cuid} is a cloned experiment (cloned experiments cannot be resumed).",
                )
            elif e.resp.status_code == 403 and e.resp.reason == "Forbidden":
                # 权限不足
                raise ValueError(f"Project permission denied: {self.projname}")
            elif e.resp.status_code == 404 and e.resp.reason == "Not Found":
                # 传入的项目不存在
                raise ValueError(f"Project {self.projname} not found")
            elif e.resp.status_code == 404 and e.resp.reason == "Disabled Resource":
                # 传入的实验被删除
                raise ValueError(f"Experiment {cuid} has been deleted")
            elif e.resp.status_code == 409 and e.resp.reason == "Conflict":
                # 传入 cuid 但是实验不属于当前项目
                raise ValueError(f"Experiment with CUID {cuid} does not belong to project {self.projname}")
            raise e
        # 200代表实验已存在，开启更新模式
        # 201代表实验不存在，新建实验
        new = resp.status_code == 201
        # 这部分信息暂时没有用到
        self.__exp = ExperimentInfo(data)
        # 获取cos信息
        self.__get_cos()
        # 重置挂起状态
        self.pending = False
        return new

    def update_state(self, success: bool, finished_at: str = None):
        """
        更新实验状态
        :param success: 实验是否成功
        :param finished_at: 实验结束时间，格式为 ISO 8601，如果不提供则使用当前时间
        """
        put_data = {
            "state": "FINISHED" if success else "CRASHED",
            "finishedAt": finished_at,
            "from": "sdk",
        }
        put_data = {k: v for k, v in put_data.items() if v is not None}  # 移除值为None的键
        self.put(f"/project/{self.groupname}/{self.projname}/runs/{self.exp_id}/state", put_data)
        self.pending = True


client: Optional["Client"] = None
"""
一个进程只有一个客户端对象
"""


def create_client(login_info: auth.LoginInfo) -> Client:
    """
    创建客户端对象
    """
    global client
    client = Client(login_info)
    return client


def get_client() -> Client:
    """
    获取客户端对象
    :return: client
    """
    global client
    if client is None:
        raise ValueError("client object is not initialized")
    return client


def reset_client():
    """
    重置client对象
    """
    global client
    client = None


def sync_error_handler(func):
    """
    在一些接口中我们不希望线程奔溃，而是返回一个错误对象
    """

    def wrapper(*args, **kwargs) -> Tuple[Optional[Union[dict, str]], Optional[Exception]]:
        try:
            # 在装饰器中调用被装饰的异步函数
            result = func(*args, **kwargs)
            return result, None
        except requests.exceptions.Timeout:
            return None, NetworkError()
        except requests.exceptions.ConnectionError:
            return None, NetworkError()
        except Exception as e:
            return None, e

    return wrapper


__all__ = [
    "get_client",
    "reset_client",
    "create_client",
    "sync_error_handler",
    "decode_response",
    "CosClient",
    "Client",
]



"""
@author: cunyue
@file: local.py
@time: 2025/2/13 14:59
@description: 本地模式下的回调
local模式（目前）将自动调用swanboard，如果不存在则报错
"""

import random

from swanlab.env import SwanLabEnv
from swanlab.log.type import LogData
from swanlab.toolkit import ColumnInfo
from . import utils
from .. import namer as N
from ..namer import generate_colors
from ..run import get_run
from ..store import get_run_store
from ...log import swanlog

try:
    # noinspection PyPackageRequirements
    import swanboard
except ImportError:
    raise ImportError("Please install swanboard to use 'local' mode: pip install 'swanlab[dashboard]'")

from importlib.metadata import version

package_version = version("swanboard")
if package_version != "0.1.8b1":
    raise ImportError(
        "Your swanboard version does not match, please use this command to install the matching version: pip install 'swanlab[dashboard]'"
    )


import json
import os
from datetime import datetime
from typing import Tuple, Optional, TextIO
from swanlab.toolkit import RuntimeInfo, MetricInfo
from swanlab.data.callbacker.callback import SwanLabRunCallback


class LocalRunCallback(SwanLabRunCallback):

    def __init__(self) -> None:
        super().__init__()
        self.board = swanboard.SwanBoardCallback()
        # 当前日志写入文件的句柄
        self.file: Optional[TextIO] = None

    def __str__(self):
        return "SwanLabLocalRunCallback"

    def _terminal_handler(self, log_data: LogData):
        log_name = f"{datetime.now().strftime('%Y-%m-%d')}.log"
        if self.file is None:
            # 如果句柄不存在，则创建
            self.file = open(os.path.join(self.run_store.console_dir, log_name), "a", encoding="utf-8")
        elif os.path.basename(self.file.name) != log_name:
            # 如果句柄存在，但是文件名不一样，则关闭句柄，重新打开
            self.file.close()
            self.file = open(os.path.join(self.run_store.console_dir, log_name), "a", encoding="utf-8")
        # 写入日志
        for content in log_data["contents"]:
            self.file.write(content['message'] + '\n')
            self.file.flush()
        self.porter.trace_log(log_data)

    def on_init(self, proj_name: str, workspace: str, public: bool = None, logdir: str = None, *args, **kwargs):
        if logdir is not None:
            os.environ[SwanLabEnv.SWANLOG_FOLDER.value] = logdir
        self.board.on_init(proj_name)
        run_store = self.run_store
        run_store.project = proj_name
        run_store.workspace = workspace
        run_store.visibility = public
        run_store.tags = [] if run_store.tags is None else run_store.tags
        exp_count = random.randint(0, 20)
        run_store.run_name = N.generate_name(exp_count) if run_store.run_name is None else run_store.run_name
        run_store.run_colors = generate_colors(random.randint(0, 20))
        run_store.run_id = N.generate_run_id()
        run_store.new = True

    def before_init_experiment(
        self,
        run_id: str,
        exp_name: str,
        description: str,
        colors: Tuple[str, str],
        *args,
        **kwargs,
    ):
        #  FIXME num 在 dashboard 中被要求传递但是没用上 🤡
        self.board.before_init_experiment(
            os.path.basename(self.run_store.run_dir), exp_name, description, colors=colors, num=1
        )

    def on_run(self, *args, **kwargs):
        self.porter.open_for_trace(backend='none', sync=True)
        self._start_terminal_proxy(handler=self._terminal_handler)
        self._register_sys_callback()
        utils.print_train_begin(self.run_store.run_dir)
        utils.print_watch(self.run_store.swanlog_dir)

    def on_runtime_info_update(self, r: RuntimeInfo, *args, **kwargs):
        # 更新运行时信息
        self.porter.trace_runtime_info(r)

    def on_log(self, *args, **kwargs):
        self.board.on_log(*args, **kwargs)

    def on_column_create(self, column_info: ColumnInfo, *args, **kwargs):
        self.porter.trace_column(column_info)
        # 屏蔽 board 不支持的图表类型和列类型
        if column_info.chart_type.value.chart_type not in ["line", "image", "audio", "text"]:
            return
        if column_info.cls != "CUSTOM":
            return
        self.board.on_column_create(column_info)

    def on_metric_create(self, metric_info: MetricInfo, *args, **kwargs):
        # 对于指标保存，可以随意保存，因为这里与 dashboard 没有直接交互
        # 出现任何错误直接返回
        if metric_info.error:
            return
        # ---------------------------------- 保存指标数据 ----------------------------------
        os.makedirs(os.path.dirname(metric_info.metric_file_path), exist_ok=True)
        os.makedirs(os.path.dirname(metric_info.summary_file_path), exist_ok=True)
        with open(metric_info.summary_file_path, "w+", encoding="utf-8") as f:
            f.write(json.dumps(metric_info.metric_summary, ensure_ascii=False))
        with open(metric_info.metric_file_path, "a", encoding="utf-8") as f:
            f.write(json.dumps(metric_info.metric, ensure_ascii=False) + "\n")
        # ---------------------------------- 保存媒体字节流数据 ----------------------------------
        self.porter.trace_metric(metric_info)

    def on_stop(self, error: str = None, *args, **kwargs):
        """
        训练结束，取消系统回调
        此函数被`run.finish`调用
        """

        # 写入错误信息
        if error is not None:
            with open(os.path.join(get_run_store().console_dir, "error.log"), "a") as fError:
                print(datetime.now(), file=fError)
                print(error, file=fError)
        self.board.on_stop(error)
        # 打印信息
        utils.print_watch(self.run_store.run_dir)
        self._unregister_sys_callback()
        self.porter.close_trace(success=get_run().success, error=error, epoch=swanlog.epoch + 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2025/5/8 14:36
@File: types.py
@IDE: pycharm
@Description:
    OpenAPI 相关数据结构
"""

from typing import Dict, Generic, List, TypeVar

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    def __getitem__(self, key):
        return getattr(self, key)


class Experiment(BaseModel):
    cuid: str               # 实验CUID, 唯一标识符
    name: str               # 实验名
    description: str = ""   # 实验描述
    state: str              # 实验状态, 'FINISHED' 或 'RUNNING'
    show: bool              # 显示状态
    createdAt: str          # e.g., '2024-11-23T12:28:04.286Z'
    finishedAt: str = ""    # e.g., '2024-11-23T12:28:04.286Z'
    user: Dict[str, str]    # 实验创建者, 包含 'username' 与 'name'
    profile: Dict           # 实验相关配置


class Project(BaseModel):
    cuid: str                   # 项目CUID, 唯一标识符
    name: str                   # 项目名
    description: str = ""       # 项目描述
    visibility: str             # 可见性, 'PUBLIC' 或 'PRIVATE'
    createdAt: str              # e.g., '2024-11-23T12:28:04.286Z'
    updatedAt: str              # e.g., '2024-11-23T12:28:04.286Z'
    group: Dict[str, str]       # 工作空间信息, 包含 'type', 'username', 'name'
    count: Dict[str, int] = {}  # 项目的统计信息


D = TypeVar("D")


class ApiResponse(BaseModel, Generic[D]):
    code: int    # HTTP状态码
    errmsg: str  # API错误消息, 只有请求错误时非空
    data: D      # 返回数据


class Pagination(BaseModel, Generic[D]):
    total: int      # 总数
    list: List[D]   # 列表数据，泛型



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-04-26 17:06:22
@File: swanlab\integration\zhipuai\__init__.py
@IDE: vscode
@Description:
    ZhipuAI 集成
"""
from .zhipuai import autolog



"""
@author: cunyue
@file: test_client.py
@time: 2025/6/16 15:51
@description: 测试客户端功能
"""

import os

import nanoid
import pytest
import requests_mock
import responses
from responses import registries

from swanlab.core_python import create_client, Client, CosClient, reset_client
from swanlab.core_python.auth import login_by_key
from swanlab.package import get_host_api
from swanlab.toolkit import MediaBuffer
from tutils import is_skip_cloud_test, TEMP_PATH, API_KEY
from tutils.setup import *


# ---------------------------------- mock 请求工具函数 ----------------------------------


class UseMocker(requests_mock.Mocker):
    """
    使用request_mock库进行mock测试，由于现在绝大部分请求都在get_host_api上，所以封装一层
    """

    def __init__(self, base_url: str = None):
        super().__init__()
        base_url = base_url or get_host_api()
        self.base_url = base_url

    def get(self, router, *args, **kwargs):
        return super().get(*(self.base_url + router, *args), **kwargs)

    def post(self, router, *args, **kwargs):
        return super().post(*(self.base_url + router, *args), **kwargs)

    def put(self, router, *args, **kwargs):
        return super().put(*(self.base_url + router, *args), **kwargs)

    def patch(self, router, *args, **kwargs):
        return super().patch(*(self.base_url + router, *args), **kwargs)

    def delete(self, router, *args, **kwargs):
        return super().delete(*(self.base_url + router, *args), **kwargs)


def test_use_mocker():
    with UseMocker() as m:
        m.post("/tmp", text="mock")
        import requests
        from swanlab.package import get_host_api

        resp = requests.post(get_host_api() + "/tmp")
        assert resp.text == "mock"


# ---------------------------------- 测试客户端 ----------------------------------


def test_decode_response():
    with UseMocker() as mocker:
        mocker.post("/json", json={"test": "test"})
        mocker.post("/text", text="test")
        with UseMockRunState() as run_state:
            client = run_state.client
            data, _ = client.post("/json")
            assert data == {"test": "test"}
            data, _ = client.post("/text")
            assert data == "test"


@responses.activate(registry=registries.OrderedRegistry)
def test_retry():
    """
    测试重试机制
    """
    from swanlab.package import get_host_api

    url = get_host_api() + "/retry"
    rsp1 = responses.get(url, body="Error", status=500)
    rsp2 = responses.get(url, body="Error", status=500)
    rsp3 = responses.get(url, body="Error", status=500)
    rsp4 = responses.get(url, body="OK", status=200)
    with UseMockRunState() as run_state:
        client = run_state.client
        data, _ = client.get("/retry")
        assert data == "OK"
        assert rsp1.call_count == 1
        assert rsp2.call_count == 1
        assert rsp3.call_count == 1
        assert rsp4.call_count == 1


@pytest.mark.skipif(is_skip_cloud_test, reason="skip cloud test")
class TestCosSuite:
    http: Client = None
    alphabet = "abcdefghijklmnopqrstuvwxyz"
    project_name = nanoid.generate(alphabet)
    experiment_name = nanoid.generate(alphabet)
    file_path = os.path.join(TEMP_PATH, nanoid.generate(alphabet))
    now_refresh_time = 1
    pre_refresh_time = CosClient.REFRESH_TIME

    @classmethod
    def setup_class(cls):
        CosClient.REFRESH_TIME = cls.now_refresh_time
        # 这里不测试保存token的功能
        login_info = login_by_key(API_KEY, save=False)
        cls.http = create_client(login_info)
        cls.http.mount_project(cls.project_name)
        cls.http.mount_exp(cls.experiment_name, ('#ffffff', '#ffffff'))
        # temp路径写一个文件上传
        with open(cls.file_path, "w") as f:
            f.write("test")

    @classmethod
    def teardown_class(cls):
        CosClient.REFRESH_TIME = cls.pre_refresh_time

    def test_cos_ok(self):
        assert self.http is not None
        assert self.http.cos is not None

    def test_cos_upload(self):
        # 新建一个文件对象
        buffer = MediaBuffer()
        buffer.write(b"test")
        buffer.file_name = "test"
        self.http.upload(buffer)
        # 为了开发方便，测试刷新功能关闭

        # # 开发版本设置的过期时间为3s，等待过期
        # time.sleep(3)
        # # 重新上传，测试刷新
        # assert self.http.cos.should_refresh is True
        # self.http.upload(buffer)


@pytest.mark.skipif(is_skip_cloud_test, reason="skip cloud test")
class TestExpSuite:
    """
    测试实验相关的功能
    """

    client: Client = None

    def setup_method(self):
        """
        初始化客户端对象
        """
        login_info = login_by_key(API_KEY, save=False)
        self.client = create_client(login_info)

    @staticmethod
    def teardown_method():
        """
        清理实验
        """
        reset_client()

    def test_create_exp_before_create_project(self):
        """
        实验创建前未创建项目，应该抛出异常
        """
        with pytest.raises(NotImplementedError) as e:
            self.client.mount_exp(nanoid.generate(), ('#ffffff', '#ffffff'))
        assert "Project not mounted, please call mount_project() first" == str(
            e.value
        ), "Expected ValueError when creating experiment without project"

    def test_create_exp_with_cuid(self):
        """
        测试创建实验时使用cuid
        cuid 必须为21位字符串且为小写字母和数字组成
        """
        cuid = nanoid.generate(alphabet="abcdefghijklmnopqrstuvwxyz0123456789", size=21)
        exp_name = nanoid.generate()
        self.client.mount_project(nanoid.generate())
        self.client.mount_exp(exp_name, ('#ffffff', '#ffffff'), cuid=cuid)
        assert self.client.exp_id == cuid, "Experiment ID should match the provided cuid"

    def test_exp_must_exist(self):
        """
        测试实验在 cuid 传递时实验必须存在
        """
        proj_name = nanoid.generate()
        exp_name = nanoid.generate()
        self.client.mount_project(proj_name)
        cuid = nanoid.generate(alphabet="abcdefghijklmnopqrstuvwxyz0123456789", size=21)
        with pytest.raises(AssertionError):
            # 如果传递 must_exist=True，则 cuid 必须传递
            self.client.mount_exp(exp_name, ('#ffffff', '#ffffff'), must_exist=True)
        with pytest.raises(RuntimeError):
            self.client.mount_exp(exp_name, ('#ffffff', '#ffffff'), cuid=cuid, must_exist=True)

    def test_exp_is_exist(self):
        """
        测试实验是否存在
        """
        proj_name = nanoid.generate()
        exp_name = "test"
        cuid = nanoid.generate(alphabet="abcdefghijklmnopqrstuvwxyz0123456789", size=21)
        self.client.mount_project(proj_name)
        new = self.client.mount_exp(exp_name, ('#ffffff', '#ffffff'), cuid=cuid)
        assert new is True, "Experiment should not exist before creation"
        new = self.client.mount_exp(exp_name, ('#ffffff', '#ffffff'), cuid=cuid)
        assert new is False, "Experiment should exist after creation"



import pytest

from swanlab.data.run.metadata.hardware.gpu.metax import MetaxCollector, map_metax_gpu

try:
    driver, maca_version, gpu_map = map_metax_gpu()
    max_mem_value = 0
    for gpu_id in gpu_map:
        mem_value = gpu_map[gpu_id]["memory"]
        if mem_value > max_mem_value:
            max_mem_value = mem_value
    max_mem_value = max_mem_value * 1024
except Exception as e:
    driver = None
    maca_version = None
    gpu_map = None
    max_mem_value = None


@pytest.mark.skipif(maca_version is None, reason="MetaX GPU not found")
def test_metax_gpu_info():
    collector = MetaxCollector(gpu_map, max_mem_value)
    collector()
    assert collector.collect_num == 1



import random

from transformers import Qwen2Config, Qwen2ForCausalLM, TrainingArguments, Trainer
from datasets import Dataset
import torch

config = Qwen2Config(
    vocab_size=24,  # GPT-2 标准词汇表大小
    hidden_size=16,
    intermediate_size=32,
    num_hidden_layers=3,
    num_attention_heads=4,
    num_key_value_heads=4,
)

model = Qwen2ForCausalLM(config)


seq = torch.randint(0, 24, (50, 32)).tolist()
ds = Dataset.from_dict({"sequence": seq})
ds = ds.train_test_split(0.2)

args = TrainingArguments(
    output_dir="./test_transformers_output",
    max_steps=120,
    per_device_train_batch_size=2,
    report_to="swanlab",
    logging_steps=1,
    remove_unused_columns=False,
)


def collator(examples):
    data = [example["sequence"] for example in examples]
    data = torch.tensor(data, dtype=torch.int)
    return {
        "input_ids": data,
        "attention_mask": torch.ones_like(data),
        "labels": data,
    }


trainer = Trainer(model, args, data_collator=collator, train_dataset=ds["train"], eval_dataset=ds["test"])

trainer.train()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/6/3 15:39
@File: wrapper.py
@IDE: pycharm
@Description:
    包装器
"""
from typing import Union, List, Optional

from swanlab.error import DataTypeError
from swanlab.toolkit import ParseResult, ParseErrorInfo, MediaType, ChartReference
from .custom_charts import Echarts
from .line import Line


class DataWrapper:
    """
    数据包装器，负责解析数据，返回相应的数据类型

    一个数据包装器对应swanlab.log字典内的一个key的值
    """

    def __init__(
        self,
        key: str,
        data: Union[List[Line], List[MediaType]],
        reference: ChartReference = "STEP",
    ):
        """
        :param key: key名称，为编码前的名称
        :param data: log的数据
        """
        self.__key = key
        self.__data = data
        self.__error = None
        self.__result: Optional[ParseResult] = None
        self.__reference = reference
        # 保证data内所有数据类型相同，否则返回TypeError
        t = self.__data[0].__class__
        for i in self.__data:
            if i.__class__ != t:
                raise TypeError(f"All data in DataWrapper must be the same type, got {t} and {i.__class__}")

    @property
    def is_line(self) -> bool:
        """
        是否是Line类型
        """
        return isinstance(self.__data[0], Line) and len(self.__data) == 1

    @property
    def is_custom(self) -> bool:
        """
        是否是自定义类型（比如 ECharts）
        """
        return all(isinstance(i, Echarts) for i in self.__data)

    @property
    def type(self):
        """
        数据类型
        """
        return self.__data[0].__class__

    @property
    def parsed(self) -> bool:
        """
        是否已经解析过数据
        """
        return self.__result is not None or self.__error is not None

    @property
    def error(self) -> Optional[ParseErrorInfo]:
        """
        解析时候的错误信息
        """
        return self.__error

    def get_class(self):
        """
        获取数据类型的类名
        """
        return self.__data[0].__class__

    def parse(self, **kwargs) -> Optional[ParseResult]:
        """
        将数据解析成对应的数据类型
        :param kwargs: 传入的参数,必须包含step
        """
        if self.parsed:
            return self.__result
        result = ParseResult(reference=self.__reference)
        result.step = kwargs["step"]  # 挂载step
        d = self.__data[0]
        result.section = d.get_section()
        result.chart = d.get_chart()
        # ---------------------------------- 特别处理Line类型 ----------------------------------

        # [Line]
        if self.type == Line:
            if len(self.__data) > 1:
                self.__error = ParseErrorInfo("float", "list(Line)", result.chart)
            else:
                d.inject(**kwargs)
                try:
                    result.float, _ = d.parse()
                except DataTypeError as e:
                    self.__error = ParseErrorInfo(e.expected, e.got, result.chart)
            self.__result = result
            return self.__result

        # ---------------------------------- 处理其他类型 ----------------------------------
        # [MediaType]
        data, buffers, more = [], [], []
        for i in self.__data:
            try:
                i.inject(**kwargs)
                d, r = i.parse()
            except DataTypeError as e:
                self.__error = ParseErrorInfo(e.expected, e.got, result.chart)
                return None
            data.append(d)
            buffers.append(r)
            more.append(i.get_more())
        result.strings = data
        # 过滤掉空列表
        result.buffers = self.__filter_list(buffers)
        result.more = self.__filter_list(more)
        self.__result = result
        return self.__result

    @staticmethod
    def __filter_list(li: List):
        """
        如果li长度大于0且如果l内部不全是None，返回l，否则返回None
        """
        if len(li) > 0 and any(i is not None for i in li):
            return li
        return None

    @classmethod
    def create_duplicate_error(cls) -> ParseErrorInfo:
        """
        快捷创建一个重复错误
        """
        return ParseErrorInfo(None, None, None, duplicated=True)
